		/* Auxiliary lexical routines  */


#include <cstdio>
#include <iostream.h>
#include "utils.h"
#include "AST.h"
#include "tokens.h"

/* The line number on which character N of the input apppears. */
size_t File::posnToLineNumber (unsigned long N)
{
    unsigned int i;
    if (N == 0)
	return 0;
    for (i = 0; i < lineEndings.size(); i += 1)
	if (N <= lineEndings[i])
	    return i;
    return i;
}

/* Record that N is the position of a newline in the source program */
/* file.  It is assumed that this function is called with increasing */
/* values of N. */
void File::defineLineEnd (unsigned long N)
{
    lineEndings.push_back (N);
}

static int errorCount = 0;

/* Print a position in a format suitable for errors or warnings. */
ostream & operator << ( ostream &os, const SourcePosn &position )
{
    if (position.unknown())
	return os << "?: ";
    else 
	return os << *position.file->name() << ":"
		  << position.posnToLineNumber() << ": ";
}

/* Report message at source position N, giving MSG as the text. */
void Message (const string& msg, const SourcePosn N = NoSourcePosition)
{
    cerr << N << msg << endl;
    cerr.flush();
}

/* Report error at source position N, giving MSG as the text of the */
/* error message. */
void Error (const string& msg, const SourcePosn N = NoSourcePosition)
{
    errorCount += 1;
    Message (msg + '.', N);
}

/* Report warning at source position N, giving MSG as the text of the */
/* error message. */
void Warning (const string& msg, SourcePosn N = NoSourcePosition)
{
    Message ("warning: " + msg + '.', N );
}

/* The number of times Error has been called. */
extern int NumErrors()
{
    return errorCount; 
}

/* An array type with given ELEMENTTYPE and DIMS dimensions.  When */
/* DIMS=0, equals ELEMENTTYPE. */
TypeNode* makeArrayType (TypeNode* elementType, int dims)
{
    while (dims > 0) {
	elementType = new JavaArrayTypeNode (elementType);
	dims -= 1;
    }
    return elementType;
}

struct InternTableEntry {
    InternTableEntry(string* datum0, InternTableEntry* next0)
	: datum(datum0), next(next0) {}
    InternTableEntry(const char* datum0, InternTableEntry* next0)
	: datum(new string(datum0)), next(next0) {}
    string* datum;
    InternTableEntry* next;
};

static InternTableEntry** internTable;
static int internTableSize = 0;

static unsigned int pjw_hash (const char* x, size_t n, size_t k)
{
    unsigned int r;
    const char* lim = x+n;

    for (r = 0; x != lim; x += 1) {
	r = (r << 4) + *x;
	if (r > 0x0fffffff) {
	    r ^= (r >> 24) & 0xf0;
	    r &= 0x0fffffff;
	}
    }
    return r % k;
}

static void createInternTable()
{
    if (internTable == NULL) {
	internTableSize = 1019;
	internTable = new InternTableEntry*[internTableSize];
	for (int i = 0; i < internTableSize; i++)
	  internTable[i] = NULL;
    }
}

/* The unique string* containing the current contents of TEXT. */
const string* intern (const char* text, size_t len)
{
    createInternTable();

    unsigned int h = pjw_hash(text, len, internTableSize);

    InternTableEntry** p;
    for (p = &internTable[h]; *p != NULL; p = &(*p)->next) {
	if (text == *(*p)->datum) 
	    return (*p)->datum;
    }
    *p = new InternTableEntry(new string(text, len), *p);
    return (*p)->datum;
}

/* The unique string* containing the current contents of TEXT. */
const string* intern (const char* text)
{
    return intern(text, strlen(text));
}

/* The unique string* containing the current contents of TEXT. */
const string* intern (const string& text)
{
    return intern(text.c_str(), text.size());
}

/* Parse array access/array type stuff */

TreeNode *TreeNode::asExpr()
{
  unimplemented("asExpr");
  return NULL;
}

TreeNode *TreeNode::asExpr2(TreeNode *)
{
  unimplemented("asExpr2");
  return NULL;
}

TypeNode *TreeNode::asType()
{
  unimplemented("asType");
  return NULL;
}

static void noQualifiers(TreeNode *t)
{
  if (t->flags())
    Error("Unexpected qualifiers " + stringifyModifiers(t->flags()),
	  t->position());
}

TreeNode *ArrayNameNode::asExpr()
{
  TreeNode *base = new ObjectNode(name());
  TreeNode *aspecs = arrayspecifiers();
  int a = aspecs->arity();

  noQualifiers(this);

  for (int i = 0; i < a; )
    base = aspecs->child(i++)->asExpr2(base);

  return base;
}

TypeNode *ArrayNameNode::asType()
{
  TypeNode *base = new TypeNameNode(name());
  TreeNode *aspecs = arrayspecifiers();

  base->modifiers(flags());

  for (int i = aspecs->arity(); i > 0; )
    base = aspecs->child(--i)->asType2(base);

  return base;
}

TypeNode *arraySpecifiers2Type(llist<TreeNode *> *aspecs, TypeNode *base)
{
  foreach (aspec, llist<TreeNode *>, *dreverse(aspecs))
    base = (*aspec)->asType2(base);

  return base;
}

TreeNode *EmptyArrayNode::asExpr2(TreeNode *array)
{
  Error("Missing index expression", position());
  return array;
}

TypeNode *EmptyArrayNode::asType2(TypeNode *base)
{
  TypeNode *t = new JavaArrayTypeNode(base);
  t->modifiers(flags());
  return t;
}

TreeNode *ExpressionArrayNode::asExpr2(TreeNode *array)
{
  noQualifiers(this);
  return new ArrayAccessNode(array, expr());
}

TypeNode *ExpressionArrayNode::asType2(TypeNode *base)
{
  /* This is an error, except if the expression is:
     - a double literal of the form nd (n a positive integer)
     - (id.id...)d (a cast) */
  TypeNode *t = expr()->asType2(base);
  t->modifiers(flags());
  return t;
}

TypeNode *TreeNode::asType2(TypeNode *base)
{
  Error("type expected", position());
  return base;
}

TypeNode *PointNode::asType2(TypeNode *base)
{
  if (args()->arity() != 1)
    return TreeNode::asType2(base);
  else
    return args()->child(0)->asType2(base);
}

TypeNode *PrimitiveLitNode::asType2(TypeNode *base)
{
  Literal l = literal();

  if (l.isNd())
    // Note that integral doubles are represented exactly...
    return new TitaniumArrayTypeNode
      (base, new PrimitiveLitNode((int32)l.doubleValue()));

  return TreeNode::asType2(base);
}

bool TreeNode::isd() const
{
  return false;
}

bool ObjectNode::isd() const
{
  TreeNode *n = name();

  return n->qualifier()->absent() && *n->ident() == "d";
}

bool TypeNameNode::isNamedType() const
{
  return true;
}

TypeNode *CastNode::asType2(TypeNode *base)
{
  TypeNode *dt = dtype();

  if (opnd0()->isd() && dt->isNamedType() && !dt->modifiers())
    return new TitaniumArrayTypeNode(base, new ObjectNode(dt->name()));
  return TreeNode::asType2(base);
}

TreeNode *TitaniumArrayNode::asExpr2(TreeNode *array)
{
  Error("unexpected identifier 'd'", position());
  return new ArrayAccessNode(array, new PointNode(cons(expr())));
}

TypeNode *TitaniumArrayNode::asType2(TypeNode *base)
{
  TypeNode *t = new TitaniumArrayTypeNode(base, expr());
  t->modifiers(flags());
  return t;
}

llist<TreeNode*>* makeVarDeclNodes(bool isfinal, TypeNode *type,
				   llist<DeclaratorTuple> *decls)
{
  llist<TreeNode *> *dlist = NULL;

  foreach (decl, llist<DeclaratorTuple>, *decls) {
    dlist = cons(static_cast<TreeNode*>
		 (new VarDeclNode(isfinal, makeArrayType(type, (*decl).dims),
				  (*decl).name, (*decl).initExpr)),
		 dlist);
  }
  free_all(decls);

  return dlist;
}
