(require 'cl)

(unless (fboundp 'string-prefix-p)
  ;; 24 Oct 1997 gnu.emacs.help
  ;; Philippe Schnoebelen <phs@mirabelle.lsv.ens-cachan.fr>
  (defun string-prefix-p (s1 s2)
    "True if string S1 is a prefix of S2 (i.e. S2 starts with S1)"
    (equal 0 (string-match (regexp-quote s1) s2))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *methods* nil
  "Association list of method names")

(defvar *attributes* nil
  "Association list of attribute names")

(defvar *states* nil
  "Association list of state variable names")

(defvar *children* nil
  "Association list of child nodes' names")

(defvar *lists* nil
  "Association list of nodes' names representing lists")

(defvar *superclasses* nil
  "Association list of nodes' immediate superclasses")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro defmethod (name defn)
  (` (push (list '(, name) '(, defn)) *methods*)))

(defmacro defattr (name type value)
  (` (progn
       (push (list '(, name) '(, type) '(, value)) *attributes*)
       (defattr* '(, name) '(, type) '(, value)))))

(defmacro defattr-default (name type value)
  (` (progn
       (push (list '(, name) '(, type) '(, value)) *attributes*)
       (defattr-default* '(, name) '(, type) '(, value)))))

(defmacro defstate (name type &optional initializer)
  (` (push (list '(, name) '(, type) '(, initializer)) *states*)))

(defmacro defchild (name type)
  (` (progn
       (push (list '(, name) '(, type)) *children*)
       (defchild* '(, name) '(, type)))))

(defmacro deflist (name element)
  (` (progn
       (push (list '(, name) '(, element)) *lists*)
       (deflist* '(, name) '(, element)))))

(defmacro defnode (name parents fields comment &optional methods states)
  (` (progn
       (check-node '(, name) '(, fields) '(, methods) '(, states))
       (push (cons '(, name) '(, parents)) *superclasses*)
       (defnode* '(, name) '(, parents) '(, fields) '(, methods) '(, states)))))

(defmacro defbasenode (name parents produce-decls-p comment &optional methods states)
  (` (progn
       (and (, produce-decls-p) (check-node '(, name) () '(, methods) '(, states)))
       (push (cons '(, name) '(, parents)) *superclasses*)
       (defbasenode* '(, name) '(, parents) '(, produce-decls-p) '(, methods) '(, states)))))

(defmacro deflistnode (name comment methods)
  (` (progn
       (check-methods '(, name) '(, methods))
       (push (list '(, name) '(TreeListNode)) *superclasses*)
       (deflistnode* '(, name) '(, methods)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(mapc (function (lambda (handler)
		  (defalias handler 'ignore)))
      '(defattr* defattr-default* defchild* deflist*
	 defnode* defbasenode* deflistnode*))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defalias 'finalize-buffer 'ignore)
(defalias 'initialize-buffer 'insert-c++-comment-prefix)

(defun insert-c++-comment-prefix (nodes builder)
  (insert "// Do not edit this file!\n"
	  "//\n"
	  "// Automatically generated by Emacs:\n"
	  "//\n"
	  "//    node defs:  " (file-name-nondirectory nodes) "\n"
	  "//    builder  :  " (file-name-nondirectory builder) "\n"
	  "//    timestamp:  " (current-time-string) "\n"
	  "\n"
	  "\n"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun check-membership (node-name key-name keys registry)
  "Verify that one set of a node's characteristics have been defined."
  (while keys
    (let ((key (car keys)))
      (setq keys (cdr keys))
      (assert (assq key registry) nil
	      "undefined %s \"%s\" in node \"%s\"" key-name key node-name))))

(defun check-fields (name fields)
  "Verify that all of a node's fields have been defined."
  (check-membership name "field" fields
		    (append *children* *attributes*)))

(defun check-methods (name methods)
  "Verify that all of a node's methods have been defined."
  (check-membership name "method" methods *methods*))

(defun check-states (name states)
  "Verify that all of a node's states have been defined."
  (check-membership name "state" states *states*))

(defun check-node (name fields methods states)
  "Verify that all of a node's characteristics have been defined."
  (check-fields name fields)
  (check-methods name methods)
  (check-states name states))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun resolve-fields (field-names)
  "Resolve a list of FIELD-NAMES into lists of detail records.
The result is a vector of three lists:

	#[ CHILDREN ATTRIBUTES FIELDS ]

The first list contains all child detail records.  The second contains
all attribute detail records.  The third contains both children and
attributes intermixed.  All three lists preserve the order of the
original list of field names."
  (let (children
	attributes
	fields)
    (while field-names
      (let* ((field-name (pop field-names))
	     (attribute (assq field-name *attributes*)))
	(if attribute
	    (progn (push attribute attributes)
		   (push attribute fields))
	  (let ((child (assq field-name *children*)))
	    (assert child)
	    (push child children)
	    (push child fields)))))
    (vector (nreverse children)
	    (nreverse attributes)
	    (nreverse fields))))

(defun resolve-states (state-names)
  "Resolve a list of STATE-NAMES into a list of state detail records."
  (mapcar (function (lambda (state-name)
		      (assq state-name *states*)))
	  state-names))

(defun list-type (type)
  (let ((lookup (assq type *lists*)))
    (and lookup
	 (cadr lookup))))

(defun to-c++-type (type)
  (let ((ltype (list-type type)))
    (cond (ltype (format "llist<%s*> *" ltype))
	  ((stringp type) type)
	  (t (format "%s *" type)))))

(defun to-c++-node-type (type)
  (if (stringp type) type
    (format "%s*" type)))

(defun any-subclass-p (children parent)
  (and children
       (or (is-subclass-p (car children) parent)
	   (any-subclass-p (cdr children) parent))))

(defun is-subclass-p (child parent)
  (or (eq child parent)
      (any-subclass-p (cdr (assq child *superclasses*)) parent)))

(defun basic-field-name (field)
  (let ((name (symbol-name field)))
    (if (string-match "\\(.*\\):" name)
	(match-string 1 name)
      name)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun eval-file (file)
  (load (expand-file-name file) nil t t))

(defun build-nodes (target nodes builder)
  (interactive "FFile to build: \nfNode definitions file: \nfBuilder: \n")
  (let ((buffer (create-file-buffer target))
	(debug-on-error (or debug-on-error
			    (interactive-p))))
    (if debug-on-error
	(switch-to-buffer buffer)
      (set-buffer buffer))
    (eval-file builder)
    (initialize-buffer nodes builder)
    (eval-file nodes)
    (finalize-buffer)
    (write-file target)
    (unless debug-on-error
      (kill-buffer buffer))))

(defun batch-build-nodes ()
  (let ((target argi)
	(nodes (pop command-line-args-left))
	(builder (concat argi ".el")))
    (assert target nil "No target given.")
    (assert nodes nil "No nodes given.")
    (build-nodes target nodes builder)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'build-nodes)

(and noninteractive
     (add-hook 'command-line-functions 'batch-build-nodes))
