#ifndef _CFG_H_
#define _CFG_H_

#include <cassert>
#include <iosfwd>
#include <set>
#include "AST.h"
#include "BufferPool.h"
#include "llist.h"

class AliasInfos;
class Bitset;


// EXPRESSION-LEVEL CFG NODES
// There are one of these nodes per AST node. They arrange the AST nodes
// into execution order. This usually means postorder. The CFG for
// an IfStmtNode, for example, will contain the test, then the true and false
// clauses down different branches, and then the IfStmtNode itself. The fact
// that the IfStmtNode itself gets executed last may seem surprising, but the
// IfStmtNode doesn't actually mean anything; it is its subexpressions that
// represent real computation. See titanium-doc/cfg.txt for all the details.

class CfgNode {
protected:
  TreeNode *ast_node;

  llist<CfgNode *> *_succ;
  llist<CfgNode *> *_pred;

  int visited;

  CfgNode *next_free;

public:
  CfgNode (TreeNode *node)
    : ast_node(node),
      _succ(0),
      _pred(0),
      visited(0),
      next_free(0),
      numPred(0),
      numSucc(0),
      bblock(0)
  {
  }

  void *operator new(size_t);      
  void operator delete(void *, size_t);

  TreeNode *astNode() { return ast_node; }
  int numSucc;
  int numPred;
  Bblock *bblock;
  void nodeStatus(int v) { visited = v; }
  int nodeStatus() { return visited; }
  void addSucc(CfgNode *n) { numSucc++; _succ = cons(n, _succ); }
  void addPred(CfgNode *n) { numPred++; _pred = cons(n, _pred); }
  void delSucc(CfgNode *n) { 
    numSucc--; 
    assert((_succ = remove(n, _succ)) != (llist<CfgNode *>*)-1 ); 
  }
  void delAllSucc() { numSucc = 0; free_all(_succ); _succ = NULL; }
  void delAllPred() { numPred = 0; free_all(_pred); _pred = NULL; }
  void delPred(CfgNode *n) { 
    numPred--; 
    assert((_pred = remove(n, _pred)) != (llist<CfgNode *>*)-1 ); 
  }
  ListIterator<CfgNode *> succIter() {
    return ListIterator<CfgNode *> (_succ);
  }
  ListIterator<CfgNode *> predIter() {
    return ListIterator<CfgNode *> (_pred);
  }
  void print(ostream &os) const;
};


// BASIC BLOCKS
// The basic blocks collect individual CfgNodes together, in case the client
// is just interested in the overall pattern of control flow and doesn't need
// to iterate through every single AST node.

class Bblock {
protected:
  llist<Bblock *> *_preds;
  llist<Bblock *> *_succs;
  llist<CfgNode *> *_cfglist;
  Bblock *next_free;
  int visited;
  Bitset *_defsOut;  // defs that reach out of the bblock.
  Bitset *_defsIn;   // union of incoming defs
  Bitset *_defs;     // defs in this bblock
  Bitset *_kills;    // defs killed by this bblock
  AliasInfos *_aliasesIn; // set of aliases at beg of bblock
  AliasInfos *_aliasesOut; // set of aliases at end of bblock
  
public:
  Bblock();

  void *operator new(size_t);      
  void operator delete(void *, size_t);

  ListIterator<Bblock *> succIter() {
    return ListIterator<Bblock *>(_succs);
  }
  ListIterator<Bblock *>predIter() {
    return ListIterator<Bblock *>(_preds);
  }
  ListIterator<CfgNode *>cfgIter() {
    return ListIterator<CfgNode *>(_cfglist);
  }
  void nodeStatus(int v) { visited = v; }
  int nodeStatus() { return visited; }
  void addSucc(Bblock *b) { push(_succs, b); }
  void addPred(Bblock *b) { push(_preds, b); }
  void addCfg(CfgNode *c) { 
    _cfglist = extend(_cfglist, cons(c));
    c->bblock = this; 
  }
  Bitset *defs() { return _defs; }
  Bitset *kills() { return _kills; }
  Bitset *defsIn() { return _defsIn; }
  Bitset *defsOut() { return _defsOut; }
  AliasInfos *aliasesIn() { return _aliasesIn; }
  AliasInfos *aliasesOut() { return _aliasesOut; }  
  void setDefs(Bitset *defs) { _defs = defs; }
  void setKills(Bitset *kills) { _kills = kills; }
  void setDefsIn(Bitset *defs) { _defsIn = defs; }
  void setDefsOut(Bitset *defs) { _defsOut = defs; }
  void setAliasesIn(AliasInfos *ai) { _aliasesIn = ai; }
  void setAliasesOut(AliasInfos *ai) { _aliasesOut = ai; }  
  void print(ostream &os);
  int number;
  int defsize;
};


////////////////////////////////////////////////////////////////////////
//
// REUSABLE NODE STORAGE
//
// Repeatedly allocating and deleting graph nodes could be expensive
// and error-prone; instead we try to reuse nodes. We do this by
// allocating nodes from BufferPools, "pseudo-heaps" that store nodes
// of one type and can be reset to free all the nodes at once. A
// collection of BufferPools is a NodeStorage. Each method contains
// one NodeStorage.  Graph node classes have custom allocation
// operators that use corresponding buffer pools for managing instance
// memory.

// One of these objects exists for every method; used for storage reuse.
class NodeStorage {
public:
  // The pseudoconstructors need to know what the current NodeStorage object
  // to get nodes from is. This variable tells them; the client needs to set
  // it to the NodeStorage object for the current method.
  static NodeStorage *current;
  
  BufferPool< sizeof( CfgNode ) > cfgHeap;
  BufferPool< sizeof( Bblock ) > bblockHeap;

  void reset() {
    cfgHeap.reset();
    bblockHeap.reset();
  }
};


inline void *CfgNode::operator new(size_t size)
{
  assert(size == sizeof(CfgNode));
  return NodeStorage::current->cfgHeap.alloc();
}

inline void CfgNode::operator delete(void *buffer, size_t size)
{
  assert(size == sizeof(CfgNode));
  NodeStorage::current->cfgHeap.release(buffer);
}

inline void *Bblock::operator new(size_t size)
{
  assert(size == sizeof(Bblock));
  return NodeStorage::current->bblockHeap.alloc();
}

inline void Bblock::operator delete(void *buffer, size_t size)
{
  assert(size == sizeof(Bblock));
  NodeStorage::current->bblockHeap.release(buffer);
}

////////////////////////////////////////////////////////////////////////

typedef set< CfgNode * > CFGset;

llist<CfgNode *> *CfgNodesWithinSubtree(const llist<CfgNode *> *l,
					TreeNode *t);
void TraverseCfg(CfgNode *c, void action(CfgNode *));
void TraverseCfgExtents(CfgNode *from, CfgNode *to, void action(CfgNode *));
void TraverseCfgScope(TreeNode *scope, void action(CfgNode *));
void UnmarkCfg(CfgNode *c);
void UnmarkCfgExts(CfgNode *from, CfgNode *to);
void UnmarkBblocks(Bblock *b);
void traverseCfgBblock(CfgNode *c, Bblock *currBblock, CfgNode *lastCfg);
void TraverseBblocksIter(Bblock *b, bool action(Bblock *b));
void TraverseBblocks(Bblock *b, void action(Bblock *b));
void TraverseBblocksScope(TreeNode *scope, void action(Bblock *b));
CfgNode *LastMethodNode(TreeNode *p);
CFGset *copy(CFGset *s);
treeSet * list_to_set(llist<CfgNode *> *l);
CFGset * list_to_CFGset(llist<CfgNode *> *l);
const CFGuset * list_to_CFGuset(CFGusetFactory &U, llist<CfgNode *> *l);
void summarizeCFG(llist<CfgNode *> * l, ostream &os);
void shortCFGset(CFGset *s, ostream &os);
void shortCFGuset(const CFGuset *s, ostream &os);
void PrintBblockAction(Bblock *b);
void setupBblocksMethod(TreeNode *method);
void setupUdMethod(TreeNode *method);


#endif // !_CFG_H_
