#include <list>
#include "AST.h"
#include "CodeContext.h"
#include "CtLocal.h"
#include "CtRegion.h"
#include "PrimitiveDecl.h"
#include "TypeDecl.h"
#include "code-assign.h"
#include "code-call.h"
#include "code-grid.h"
#include "code.h"
#include "compiler.h"
#include "config.h"
#include "decls.h"
#include "delimit.h"
#include "domain-decls.h"
#include "jni-bridge.h"
#include "osstream.h"
#include "stl-queue.h"

/* #define DEBUG_INITIALIZERS 1 */

typedef STL_QUEUE( string,     list< string     > )  StringQueue;
typedef STL_QUEUE( TypeNode *, list< TypeNode * > )  TypeQueue;


int initTempNum = 0;

static const string getRegion(TreeNode *alloc, CodeContext &context)
{
  TreeNode *region = alloc->region();
  string regexpr = region->type()->isLocal()
    ? "MyRegion((Region)"
    : "MyGRegion(";

  regexpr += region->emitExpression(context);
  regexpr += ')';
  return regexpr;
}

const string AllocateNode::emitExpression2( CodeContext &context )
{
    TypeNode &t = *dtype();
    ClassDecl &cdecl = static_cast< ClassDecl & >( *t.decl() );
    assert( cdecl.category() & (Decl::Class | Decl::Interface) );

    string ctor;
    vector< string > parameters;

    if (javaDefinedType(&t)) {
	Decl &d = *decl();
	ctor = "Ti_AllocObject";
	parameters.push_back( '"' + javaClassName(&cdecl) + '"' );
	parameters.push_back( '"' + d.type()->signature() + '"' );
    } else if (t.isPointType()) {
	ctor = "point_" + t.child(0)->typeName() + "_alloc";
    } else if (t.isRectDomainType()) {
	ctor = "rect_domain_" + t.child(0)->typeName() + "_alloc";
    } else if (t.isTitaniumArrayType()) {
// 	ctor = "tit_array_" + t->child(1)->typeName() + "_alloc";
// 	parameters.push_back( "(tit_array_" + t->child(1)->typeName()
// 			      + " <" + t->child(0)->cType() + "> *)NULL" );
	assert(0);
    } else {
	ctor = decl()->cMethodNameStatic();
	
	if (!t.isImmutable())
	    if (&cdecl == SharedRegionDecl)
		ctor = "newsregion";
	    else if (&cdecl == PrivateRegionDecl)
		ctor = "newpregion";
	    else {
		ctor = decl()->cMethodNameStatic();
		const CtType &className = cdecl.cType();
		const string instance = declareTemporary( context );
	
		if (region()->absent())
		    context << 
		      instance << " = (" << className << " *) " <<
		      "ti_malloc(sizeof(" << className << "))";
		else
		  {
		    string regionExpr = getRegion(this, context);
		    context << instance << " = (" << className << " *) "
			    << "ralloc(" << regionExpr << ", sizeof("
			    << className << "))";
		  }

		context << ';' << endCline
			<< instance << "->class_info = &"
			<< cdecl.cDescriptorName() << ';' << endCline
			<< instance << "->intf_info = "
			<< cdecl.cIntfMTName() << ';' << endCline;
	
		parameters.push_back( instance);
		context.depend( className );
	    }
    }
    
    args()->emitExpressionList( context, parameters );
    return ctor + '(' + delimit( parameters ) + ')';
}

const string callNoArgConstructor(TypeNode *t)
{
  if (t->isTitaniumArrayType())
    return callGridMethod(*t, "empty", "()");

  string className, method;

  if (t->isTitaniumBuiltinType()) {
    const string n = int2string(t->tiArity());

    if (t->isPointType()) {
      return MANGLE_TI_DOMAINS_POINT_EMPTY(+, n);
    } else if (t->isRectDomainType()) {
      return string("(") + NEW_RECTDOMAIN_EMPTY(+, n) + ")";
    } else if (t->isDomainType()) {
      return string("(") + NEW_DOMAIN_EMPTY(+, n) + ")";
    } else {
      assert(0);
      return string();
    }
  } else {
    TypeDecl *d = t->decl();
    className = d->cType();
    method = MANGLE_METH_HEAD(+, "", *d->name()) + MANGLE_METHOD_NAME_END_MARKER + className;

    vector< string > parameters;
    if (!t->isImmutable())
      parameters.push_back("(" + className + " *)"
			   " calloc(1, sizeof(" + className + "))");
    return method + '(' + delimit( parameters ) + ')';
  }
}

static string javaArrayAlloc(CodeContext &context,
			     const string& region,
			     const TypeNode &type,
			     const string& length)
{
  const string result = ExprNode::declareTemporary( context, type );
  context.depend( type.elementType()->cType() );
  
  context << "JAVA_ARRAY_ALLOC("
	  << result << ", "
	  << region << ", "
	  << length << ", "
	  << type.elementType()->cType() << ", "
	  << type.elementType()->isAtomic()
	  << ");" << endCline;
  return result;
}

static string gridAlloc(const string region, TypeNode *type, int arity,
			const string& domain)
{
  return string("_ti_") + (type->isLocal() ? "" : "global_") + "CONSTRUCT("
    + type->elementType()->cType() + ", "
    + int2string(arity)
    + ")(\"" + type->position().asString() + "\", "
         + region + ", " + domain + ')';
}

static bool requiresInitialization(TypeNode *t)
{
  return t->isTitaniumArrayType() || t->isJavaArrayType() || t->isImmutable();
}

static string arrayInitializer(const string region,
			       const string constr,
			       const string constrSize,
			       StringQueue  dimExprs,
			       TypeQueue  arrTypes,
			       TypeQueue  dimTypes,
			       LocalVars &vars,
			       CodeContext &os,
			       string &initloop);

static string initializer(const string region,
			  StringQueue  dimExprs,
			  TypeQueue  arrTypes,
			  TypeQueue  dimTypes,
			  LocalVars &vars,
			  CodeContext &os,
			  string &initloop)
{
  TypeNode *eltType = arrTypes.front()->elementType();
  if (eltType->isTitaniumArrayType()) {
    dimExprs.pop();
    dimTypes.pop();
    arrTypes.pop();
    return arrayInitializer(region,
			    gridAlloc(region,
				      eltType,
				      dimTypes.front()->tiArity(),
				      dimExprs.front()),
			    (string) MANGLE_TI_DOMAINS_RECTDOMAIN_DISPATCH(+, int2string(dimTypes.front()->tiArity()), "getNumPoints", "") + dimExprs.front() + ")",
			    dimExprs, arrTypes, dimTypes, vars, os, initloop);
  } else if (eltType->isJavaArrayType()) {
    dimExprs.pop();
    dimTypes.pop();
    arrTypes.pop();
    return arrayInitializer(region,
			    javaArrayAlloc(os, region, *eltType, dimExprs.front()),
			    dimExprs.front(), 
			    dimExprs, arrTypes, dimTypes, vars, os, initloop);
  } else if (eltType->isImmutable())
    return callNoArgConstructor(eltType);
  else {
    assert(0);
    return string();
  }
}

/* tempName is the name of a variable containing the descriptor for a
   Titanium array of type t, or a pointer to the descriptor for a Java
   array of type t.  Return a string that is a C expression for a
   local pointer to the location of the data for the array. */
static string getDataLoc(TypeNode *t, const string& tempName)
{
  return (t->isTitaniumArrayType() ?
	  (t->isLocal() ?
	   (tempName + ".A") :
	   string("TO_LOCAL(") + tempName + ".A)") :
	  (tempName + "->data"));
}

static string arrayInitializer(const string region,
			       const string constr,
			       const string constrSize,
			       StringQueue  dimExprs,
			       TypeQueue  arrTypes,
			       TypeQueue  dimTypes,
			       LocalVars &vars,
			       CodeContext &os,
			       string &initloop)
{
#if DEBUG_INITIALIZERS
  cout << "arrayInitializer(" << constr << ", " << constrSize << ")\n";
  cout << "size of dimExprs = " << dimExprs.size() << "\n";
  cout << "size of arrTypes = " << arrTypes.size() << "\n";
  cout << "size of dimTypes = " << dimTypes.size() << "\n";
  cout << "  arrTypes.front() = \n";
  arrTypes.front()->print(cout, 10);
  cout << '\n';
  StringQueue de = dimExprs;
  TypeQueue at = arrTypes;
  int i = 0;
  for (TypeQueue dt = dimTypes; !de.empty();
       dt.pop(), de.pop(), at.pop(), i++) {
    cout << "  dimExprs[" << i << "] = " << de.front() << "\n"
      << "  arrTypes[" << i << "] = \n";
    at.front()->print(cout, 10);
    cout << "\n  dimTypes[" << i << "] = \n";
    dt.front()->print(cout, 10);
    cout << "\n";
  }
#endif
    

  /*
  if (!requiresInitialization(arrTypes.front()->elementType())
      || (arrTypes.front()->elementType()->isTitaniumArrayType()
	  && dimExprs.size() <= 1))
    return constr;
    */
  if (!requiresInitialization(arrTypes.front()->elementType())
      || dimExprs.size() <= 1)
    return constr;
  else {
    string tempName = string("init") + int2string(initTempNum++);
    TypeNode *elemType = arrTypes.front()->elementType();
    static const CtLocal voidPtr( PrimitiveDecl::VoidDecl.cType() );

    vars.declare( tempName, arrTypes.front()->cType() );
    vars.declare( "cur_" + tempName, voidPtr );

    // getDataLoc() needs to access "data" field
    if (arrTypes.front()->isJavaArrayType())
      vars.depend( arrTypes.front()->decl()->cType() );

    initloop = initloop +
      "initloop_head(" + tempName + ", " +
      "cur_" + tempName + ", " +
      getDataLoc(arrTypes.front(), tempName) + ", " +
      constr + ", " + constrSize + ")\n";
    ostringstream o;
    {
      CodeContext c(os, o);
      c <<
	makeAssign("(" + elemType->cType() + " *" + ")cur_" + tempName,
		   initializer(region, dimExprs, arrTypes, dimTypes,
			       vars, c, initloop),
		   true,
		   assignKind(elemType),
		   false) + ";\n";
    }
    initloop = initloop + o.str() +
      "initloop_tail(cur_" + tempName + ", " +
      elemType->cType() + ")\n";
    
    return tempName;
  }
}

// #define DEBUG_ARRAY_ALLOC DEBUG
const string AllocateArrayNode::emitExpression2( CodeContext &context )
{
  bool tiArray = type()->isTitaniumArrayType();
  TreeNode *dt = dtype();
  TypeNode *arrType = type();
  StringQueue dimExprs;
  TypeQueue dimTypes, arrTypes;
  string initloop;
  
#if DEBUG_INITIALIZERS
  cout << position().asString() << ":\n";
#endif
  foriter (p, this->dimExprs()->allChildren(), ChildIter) {
#if DEBUG_INITIALIZERS
    cout << "arrType = \n";
    arrType->print(cout, 14);
    cout << "\ndimType = \n";
    (*p)->expr()->type()->print(cout, 14);
    cout << "\n";
#endif
    if (arrType->isTitaniumArrayType()
	&& !(*p)->expr()->type()->isRectDomainType()) break;
    arrTypes.push(arrType);
    dimExprs.push((*p)->expr()->emitExpression(context));
    dimTypes.push((*p)->expr()->type());
    arrType = arrType->elementType();
  }

  string regexpr;

  if (region()->absent())
    regexpr = "NULL";
  else
    {
      static CtLocal LocalRegion(CtRegion::singleton);
      regexpr = declareTemporary(context, LocalRegion);
      const string get = getRegion(this, context);
      context << regexpr << " = " << get << ';' << endCline;
    }

  const string ret =
    tiArray ?
    arrayInitializer(regexpr,
		     gridAlloc(regexpr,
			       type(),
			       dimTypes.front()->tiArity(),
			       dimExprs.front()),
		     (string) MANGLE_TI_DOMAINS_RECTDOMAIN_DISPATCH(+, int2string(dimTypes.front()->tiArity()), "getNumPoints", "") + dimExprs.front() + ")",
		     dimExprs,
		     arrTypes,
		     dimTypes,
		     context,
		     context,
		     initloop) :
    arrayInitializer(regexpr,
		     javaArrayAlloc(context, regexpr,
				    *type(), dimExprs.front()),
		     dimExprs.front(),
		     dimExprs,
		     arrTypes,
		     dimTypes,
		     context,
		     context,
		     initloop);
  context << initloop;
  return ret;
}
