/*
  code-defs.cc

  Performs Def/Use analysis.  The first part of the file contains utilities,
  then we have the code for uses, then the code for defs.

  */

/* 
   In this text, the term "variable" means "a thing which has an lval",
   i.e. an object, a field of an object, a field of a type, or an array
   element.

   The objects and methods manipulated herein form a package that is used to
   support def/use construction. It shouldn't be needed by an optimization
   phase; such a phase should use the def/use edges hooked up directly to
   variables. 

   A Defs object stores information about which variables are defined
   in a given section of code. In the general D-U analysis, this
   section of code is always a method; hence, every method has one
   Defs object. The purpose of a Defs object is twofold. First, it provides
   a one-to-one mapping between all the unique variable definitions that 
   occur within a method, and a set of unique integers which allow the
   definitions to be represented collectively by bitvectors. The second
   purpose of the Defs object is to calculate the kills and defs created
   by a given variable definition.
   
   Definitions are represented by DefNode objects, which are simply TreeNodes
   glued to their representative bitvector indices.
*/
   

#include <cassert>
#include <fstream>
#include <iomanip>
#include <iostream>
#include "AST.h"
#include "bitset.h"
#include "code-defs.h"
#include "code-grid.h"
#include "code-util.h"
#include "code.h"
#include "config.h"
#include "decls.h"
#include "errors.h"
#include "interface.h"
#include "utils.h"
#include "optimize.h"

extern bool debug_defs;

// Return a canonical representation of the type t with no modifiers.
static xtype type2xtype(TypeNode *t)
{
  TypeNode *c = t->newModifiers((Common::Modifiers) 0);
  return (xtype) c->typeName();
}

static const string& decl2str(Decl *d)
{
  return *d->name();
}

#define type2str(t) ((t)->cType())
#define xtype2str(t) ((string) (t))
#define field2str(f) (f)

static void dump_list(ostream &os, llist<DefNode *> *l, int indent)
{
  foreach (f, llist<DefNode *>, *l) {
    (*f)->t->print(os, indent);
    os << '\n';
  }
}

static void dump_field_to_deflist(ostream &os, const string& prefix,
			       map_string_to_deflist m, int indent)
{
  for (map_string_to_deflist::const_iterator f = m.begin();
       f != m.end();
       f++) {
    for (int j = 0; j < indent; j++) os << ' ';
    os << prefix << "(" << field2str((*f).first) << ")=\n";
    dump_list(os, (*f).second, indent + 10);
  }
}

static string plural(int n, const string& s)
{
  return (n == 1) ? "1 " + s : (int2string(n) + ' ' + s + 's');
}

void Defs::print(ostream &os, int indent)
{
  int j;

  for (j = 0; j < indent; j++) os << ' ';
  os << "varDefs:";
  j = 0;
  for (map_decl_to_deflist::const_iterator v = _varDefs.begin();
       v != _varDefs.end();
       v++) {
    os << ' ' << *((*v).first)->name();
    j++;
  }
  if (j == 0) os << " none";
  os << '\n';

  for (j = 0; j < indent; j++) os << ' ';
  os << "type_anyField(type) holds for:";
  j = 0;
  for (map_type_to_bool::const_iterator a = _type_anyField.begin();
       a != _type_anyField.end();
       a++) {
    if ((*a).second) os << ' ' << xtype2str((*a).first);
    j++;
  }
  if (j == 0) os << " none";
  os << '\n';

  for (map_type_field_to_deflist::const_iterator b = _type_field_to_deflist.begin();
       b != _type_field_to_deflist.end();
       b++) {
    dump_field_to_deflist(os, "type_field_to_deflist(" + xtype2str((*b).first) + ")",
		       (*b).second, indent);
  }

  for (map_type_to_deflist::const_iterator d = _arrayDefs.begin();
       d != _arrayDefs.end();
       d++) {
    for (j = 0; j < indent; j++) os << ' ';
    int count = (*d).second->size();
    os << "array type " << xtype2str((*d).first)
      << " has " << plural(count, "arrayDef");
    os << '\n';
  }

  os << "unknownMethods: " << (unknownMethods ? "yes" : "no") << '\n';
}

void Defs::printVardefs(ostream &os, int indent) {
  for (int j = 0; j < indent; j++) os << ' ';
  for (map_decl_to_deflist::const_iterator v = _varDefs.begin();
       v != _varDefs.end();
       v++) {
    os << (*v).first->name()->c_str() << ": ";
    dump_list(cout, (*v).second, 0);
    os << endl;
  }
}

void Defs::printBitset(Bitset *bs, ostream &os, int indent) {
  for (map_decl_to_deflist::const_iterator v = _varDefs.begin();
       v != _varDefs.end();
       v++) {
    ListIterator<DefNode *>iter = (*v).second;
    for (; !iter.isDone(); iter.next()) {
      DefNode *def = *iter;
      if (bs->test(def->index)) {
	def->t->print(os, indent);
	os << endl;
      }
    }
  }
}

Defs::~Defs() {
  for (map_decl_to_deflist::const_iterator v = _varDefs.begin();
       v != _varDefs.end();
       v++) {
    llist<DefNode *> *deflist = (*v).second;
    while (deflist) {
      delete deflist->front();
      deflist = deflist->free();
    }
  }  
}

/////////////////////////////////////////////////////////////////////////////
// Uses
/////////////////////////////////////////////////////////////////////////////

inline bool Defs::varDefsWithDecl(Decl *decl)
{
  if (debug_defs) {
    cout << "varDefsWithDecl(" << decl2str(decl) << ") -> "
	 << (varDefs(decl) != NULL) << "\n";
  }
  return (varDefs(decl) != NULL);
}

inline bool Defs::fieldDefsWithinType(TypeNode *type)
{
  if (debug_defs) {
    cout << "fieldDefsWithType(" << type2str(type) << ") -> "
	 << type_anyField(type) << "\n";
  }
  return type_anyField(type);
}

void Defs::arrayDef(TreeNode *assignment, TypeNode *t)
{
  llist<DefNode *> *&l = arrayDefs(t);
  l = cons(new DefNode(assignment, nextindex++), l);

  if (debug_defs) {
    cout << "arrayDef at " << assignment->position().asString() << ":\n";
    assignment->print(cout, 4);
    cout << '\n';
  }
}

inline bool Defs::arrayDefsWithinType(TypeNode *type)
{
  return (arrayDefs(type) != NULL);
}

bool Defs::varModified(Decl *decl)
{
  return varDefsWithDecl(decl);
}

bool Defs::varMaybeModified(Decl *decl)
{
  bool vM = varModified(decl);
  bool ret = vM || (!isLocalOrParam(decl) && unknownMethods);

  if (debug_defs) {
    cout << "The thing specified by this decl (" << decl2str(decl) << ") is "
	 << (ret ? "maybe" : "not")
	 << " modified";

    cout << "\n";
    if (true) 
      cout << "  varModifed(decl)=" << vM
	   << "\n  isLocalOrParam(decl)=" << isLocalOrParam(decl)
	   << "\n  unknownMethods=" << unknownMethods
	   << "\n";
  }
  return ret;
}

bool Defs::arrayMaybeModified(TypeNode *type) {
  return arrayDefsWithinType(type);
}

bool Defs::fieldMaybeModified(TypeNode *type, const string &field)
{
  return (type_field_to_deflist(type, field) != NULL || unknownMethods);
}

bool Defs::fieldNotModified(TypeNode *type, const string &field)
{
  return !fieldMaybeModified(type, field);
}

/////////////////////////////////////////////////////////////////////////////
// Defs
/////////////////////////////////////////////////////////////////////////////

bool ObjectNode::hasLval() { return true; }
bool FieldAccessNode::hasLval() { return true; }
bool ArrayAccessNode::hasLval() { return true; }


// Makes the given node a def of everything aliasable.
void Defs::defAllAliasable(TreeNode *t) {
  map_type_field_to_deflist::iterator fanIter; 

  for (fanIter = _type_field_to_deflist.begin(); 
       fanIter != _type_field_to_deflist.end(); 
       fanIter++) {
    map_string_to_deflist::iterator strIter;
    for (strIter = fanIter->second.begin();
	 strIter != fanIter->second.end();
	 strIter++) {
      DefNode *newDef = new DefNode(t, nextindex++);
      // Add the new DefNode to the list of defs that the call makes.
      // We will use this when calculating bitsets.
      llist<DefNode *> *&l = callDefs(t);
      l = cons(newDef, l);

      // Add the new DefNode to the list of defs for the variable it defines.
      // We will use this when connecting ud edges.
      strIter->second = cons(newDef, strIter->second);
    }
  }

  map_type_to_deflist::iterator arrIter;
  for (arrIter = _arrayDefs.begin();
       arrIter != _arrayDefs.end();
       arrIter++) {
    DefNode *newDef = new DefNode(t, nextindex++);
    llist<DefNode *> *&l = callDefs(t);
    l = cons(newDef, l);

    arrIter->second = cons(newDef, arrIter->second);
  }
}

// Takes all the unknown method calls we saved in methodCalls and makes all
// of them defs of every aliasable entity.
void Defs::addCallDefs(void) {
  ListIterator<TreeNode *> iter = methodCalls;
  for (; !iter.isDone(); iter.next()) {
    defAllAliasable(*iter);
  }
}

// If this expression defines a variable, that variable is returned.
// Otherwise the return value is NULL.
TreeNode *getDefVar(TreeNode *t) {
  if (isAssignNode(t)) {
    return t->child(0);
  }
  else if (isForEachPairNode(t)) {
    return t->simpName();
  }
  else if (isParameterNode(t)) {
    return t->simpName();
  }
  return NULL;
}

// Returns a list of all the spots where a variable is defined.
llist<DefNode *> *Defs::getAllDefs(TreeNode *t) {
  if (isObjectNode(t)) {
    if (isNameNode(t->child(0)) && t->child(0)->decl() != NULL) {
      return varDefs(t->child(0)->decl());
    }
    else assert(0);
  }
  else if (isTypeFAN(t)) {
    return type_field_to_deflist(t->ftype(), *t->simpName()->ident());
  }
  else if (isOFAN(t)) {
    return type_field_to_deflist(t->object()->type(), *t->simpName()->ident());
  }
  else if (t->isArrayAccessNode()) {
    return arrayDefs(t->array()->type());
  }
  else if (isNameNode(t)) {
    return varDefs(t->decl());
  }
  else {
    cout << "Unknown def: " << endl;
    t->print(cout,0);
    cout << endl;
    assert(0);
  }
}

// If t defines a variable, then the def and kill sets given are updated
// accordingly. These def and kill sets are "running totals" for the current
// bblock.
void Defs::merge(TreeNode *t, Bitset *defs, Bitset *kills) {
  llist<DefNode *> *deflist;
  TreeNode *defvar = NULL;

  // If this expression defines a var, get the list of existing definitions
  // for that var (which will include this one).
  if (isMethodCallNode(t)) {
    deflist = callDefs(t);
  }
  else {
    defvar = getDefVar(t);
    if (!defvar) {
      return;
    }
    deflist = getAllDefs(defvar);
  }

  if (!deflist) {
    return;
  }

  // Go through the list of definitions and kill or def each one as 
  // appropriate. For example, if the var is not aliasable, then all defs
  // of it except this one should be killed.
  ListIterator<DefNode *> iter = deflist;
  for (; !iter.isDone(); iter.next()) {
    DefNode *defnode = *iter;
    if (defnode->t == t) {
      defs->set(defnode->index);
      kills->clear(defnode->index);
    }
    else {
      // If the var is aliasable, then deflist contains the list of all the
      // defs that may also be defs of this var; we don't know for sure. The 
      // kill set is currently defined as a "must kill", so we can't kill 
      // any of them.

      // Every DefNode hooked up to a MethodCallNode ought to have the
      // MethodCallNode as its definition point. The bitset indices are what
      // differentiate two DefNodes in a MethodCallNode.
      assert(!isMethodCallNode(t));

      if (isObjectNode(defvar)) {
	kills->set(defnode->index);
      }
    }
  }
}


// If t defines a variable, then existing defs of that variable in the
// reaching set are killed and the appropriate def is set. The reaching
// set is used within a bblock to say which defs reach a particular
// TreeNode. This set is temporary and only used while walking through a
// Bblock to hook up UD edges. This routine is pretty much the same logic
// as Defs::merge. The reason why we are doing it is because we lost the
// intra-bblock def/kill info when we did Defs::merge---that method is used
// to compute inter-bblock def/kill info. Memoizing to avoid redundancy would
// be a Good Idea.
void Defs::reaching(TreeNode *t, Bitset *reaching) {
  llist<DefNode *> *deflist;
  TreeNode *defvar = NULL;

  // If this expression defines a var, get the list of existing definitions
  // for that var (which will include this one).
  if (isMethodCallNode(t)) {
    deflist = callDefs(t);
  }
  else {
    defvar = getDefVar(t);
    if (!defvar) {
      return;
    }
    deflist = getAllDefs(defvar);
  }

  if (!deflist) {
    return;
  }

  ListIterator<DefNode *> iter = deflist;
  for (; !iter.isDone(); iter.next()) {
    DefNode *defnode = *iter;
    
    if (defnode->t == t) {
      reaching->set(defnode->index);
    }
    else {
      // If the var is aliasable, then deflist contains the list of all the
      // defs that may also be defs of this var; we don't know. The 
      // kills are currently defined as "must kills", so we can't kill 
      // any of them.

      // Every DefNode hooked up to a MethodCallNode ought to have the
      // MethodCallNode as its definition point. The bitset indices are what
      // differentiate two DefNodes in a MethodCallNode.
      assert(!isMethodCallNode(t));
      
      if (isObjectNode(defvar)) {
	reaching->clear(defnode->index);
      }
    }
  }
}

// Finds the instance of the defining variable in a given defining expression,
// and hooks up a d->u edge to the given use.
void connectDuEdge(TreeNode *defExpr, TreeNode *use) {
  if (isMethodCallNode(defExpr))
    return;

  // Has use been optimized away?
  if (isSRArrayAccessNode(use->parent()) ||
      isOSRArrayAccessNode(use->parent()))
    return;

  bool debug = false;
  DEBUG_PHASE(debug = true, "usedef");
  if (debug) {
    cout << "connectDuEdge(";
    PCODE(defExpr);
    cout << ", ";
    PCODE(use);
    cout << ")" << endl;
  }

  TreeNode *defVar = getDefVar(defExpr);
  assert(defVar);
  defVar->setUses(cons(use, defVar->getUses()));
}

/* The possible values of the object if expr is an ArrayAccessNode or
   ObjectFieldAccessNode, or NULL otherwise. */
Bitset *getObjectAbstractValues(TreeNode *expr)
{
  if (isOFAN(expr))
    return expr->object()->getAbstractValues();
  if (isArrayAccessNode(expr))
    return expr->array()->getAbstractValues();
  return NULL;
}

/* The possible values modified by the given MethodCallNode, or AssignNode
   into an ArrayAccessNode or ObjectFieldAccessNode. */
Bitset *getModifiedObjectValues(TreeNode *assign)
{
  if (isAssignNode(assign))
    return getObjectAbstractValues(assign->opnd0());
  if (isMethodCallNode(assign))
    return assign->modifiesValues();
  return NULL;
}

/* assign is an AssignNode or MethodCallNode, expr is an ExprNode with an
   l-value. Returns false if the assignment cannot possibly have defined
   expr (from aliasing info). */
bool aaFilter(TreeNode *assign, TreeNode *expr)
{
  Bitset *modValues = getModifiedObjectValues(assign);
  Bitset *objectValues = getObjectAbstractValues(expr);
  if (modValues != NULL && objectValues != NULL) {
    bool r = modValues->disjoint(objectValues);
    bool debug = false;
    DEBUG_PHASE(debug = true, "usedef");
    if (r && debug) {
      assign->pseudoprint(cout, 1);
      cout << " cannot possibly have defined ";
      expr->pseudoprint(cout, 0);
      cout << endl;
    }
    return !r;
  }
  return true;
}

// Given the reaching bitset for a given use, connect u->d edges. 
void Defs::connect(TreeNode *var, Bitset *reaching) {
  llist<DefNode *> *deflist = getAllDefs(var);

  if (!deflist) {
    return;
  }
  llist<TreeNode *> *udlist = NULL;
  ListIterator<DefNode *> iter = deflist;
  for (; !iter.isDone(); iter.next()) {
    DefNode *defnode = *iter;
    if (reaching->test(defnode->index) && aaFilter(defnode->t, var)) {

#if 0
  cout << "add to defs (";
  PCODE(var);
  cout << ", ";
  PCODE(defnode->t);
  cout << ")" << endl;
#endif

      udlist = cons(defnode->t, udlist);
      connectDuEdge(defnode->t, var);
    }
  }
  var->setDefs(udlist);
}

#if 0
void Defs::classVarAssignment(TreeNode *assignment)
{
  TreeNode *lhs = assignment->child(0);
  llist<TreeNode *> *&l =
    classVarDefs(lhs->ftype(), *lhs->simpName()->ident());

  l = cons(assignment, l);

  if (debug_defs) {
    cout << "classVarDef at " << lhs->position().asString() << ":\n";
    assignment->print(cout, 4);
    cout << '\n';
  }
}
#endif

void Defs::addTypeFieldDef(TypeNode *t, const string& field,
			   TreeNode *assignment)
{
  llist<DefNode *> *&l = type_field_to_deflist(t, field);
  l = cons(new DefNode(assignment, nextindex++), l);
  type_anyField(t) = true;
  if (debug_defs) {
    cout << "fieldDef < " << t->typeName() << ", " << field << " > at "
	 << assignment->position().asString() << ":\n";
    assignment->print(cout, 4);
    cout << '\n';
  }
}  

/* Note (in this) that the variable specified by d is modified at t. */
void Defs::addVarDef(TreeNode *t, Decl *d)
{
  llist<DefNode *> *&l = varDefs(d);
  l = cons(new DefNode(t, nextindex++), l);
  if (debug_defs) {
    cout << "varDef at " << t->position().asString() << ":\n";
    t->print(cout, 4);
    cout << '\n';
  }
}

void Defs::varAssignment(TreeNode *assignment)
{
  TreeNode *lhs = assignment->child(0);
  if (isNameNode(lhs->child(0))
      && lhs->child(0)->decl() != NULL) {
    addVarDef(assignment, lhs->child(0)->decl());
  } else assert(0);
}

void Defs::assignment(TreeNode *t)
{
  TreeNode *lhs = t->child(0);
  
  if (lhs->isArrayAccessNode()) {
    arrayDef(t, lhs->array()->type());
  }
  else if (isTypeFAN(lhs))
    addTypeFieldDef(lhs->ftype(), *lhs->simpName()->ident(), t);
  else if (isOFAN(lhs))
    addTypeFieldDef(lhs->object()->type(), *lhs->simpName()->ident(), t);
  else if (isObjectNode(lhs))
    varAssignment(t);
  else {
    cerr << "Unknown type of assignment (!):\n";
    t->print(cerr, 4);
    cerr << '\n';
    Error("Defs analysis can't handle unknown type of assignment.");
    assert(0);
  }
}

Defs *TreeNode::defs()
{
  undefined("defs");
  return 0;
}

Defs *ForEachStmtNode::defs()
{
  return _defs;
}

void TreeNode::findDefs(Defs *d)
{
  foriter (p, allChildren(), ChildIter)
    (*p)->findDefs(d);
}

void VarDeclNode::findDefs(Defs *d)
{
  d->addVarDef(this, decl());
}

void ForEachPairNode::findDefs(Defs *d)
{
  d->addVarDef(this, simpName()->decl());
}

void ParameterNode::findDefs(Defs *d)
{
  d->addVarDef(this, simpName()->decl());
}

#define AssignmentDefs(Node)			\
void Node::findDefs(Defs *d)			\
{						\
  d->assignment(this);				\
  foriter (p, allChildren(), ChildIter)		\
    (*p)->findDefs(d);				\
}

AssignmentDefs(AssignNode)
/*
AssignmentDefs(BinaryArithAssignNode)
AssignmentDefs(ShiftAssignNode)
AssignmentDefs(BitwiseAssignNode)
AssignmentDefs(MethodCallAssignNode)
AssignmentDefs(IncrDecrNode)
*/

/* Do findDefs() in all children.  Set d->unknownMethods if
   the method is not known to be side effect free. */
void MethodCallNode::findDefs(Defs *d)
{
  foriter (p, allChildren(), ChildIter)
    (*p)->findDefs(d);

  if (d->unknownMethods || isSideEffectFree()) return;

  if (debug_defs) {
    cout << "`unknown' method at " << position().asString() << ":\n";
    print(cout, 4);
    cout << "\n";
  }

  d->methodCalls = cons((TreeNode *)this, d->methodCalls);
  d->unknownMethods = true;
}

