#include <map>
#include <string>
#include "utils.h"
#include "AST.h"
#include "ClassDecl.h"
#include "CodeContext.h"
#include "CtExternal.h"
#include "CtGlobal.h"
#include "CtLocal.h"
#include "PrimitiveDecl.h"
#include "code.h"
#include "code-util.h"
#include "compiler.h"
#include "decls.h"
#include "delimit.h"
#include "jni-bridge.h"
#include "lgMacro.h"
#include "UniqueId.h"

typedef map< TreeNode *, string *, less< TreeNode * > > Map;

void ThisConstructorCallNode::codeGen( CodeContext &context )
{
  context.setDebugLine(this);

    vector< string > parameters;
    parameters.push_back( MANGLE_THIS_VAR(+) );
    args()->emitExpressionList( context, parameters );

    context << decl()->cMethodNameStatic() << '('
	    << delimit( parameters )
	    << ");" << endCline;
}

/* t is code in some class.  Find the ClassDeclNode for that class. */
static TreeNode *findClassDecl(TreeNode *t)
{
  while (!isClassDeclNode(t))
    t = t->parent();

  return t;
}

void SuperConstructorCallNode::codeGen( CodeContext &context )
{
    context.setDebugLine(this);

    /* Figure out what class we are in. */
    ClassDecl * const caller = static_cast< ClassDecl * >
      (findClassDecl(this)->decl());

    assert( caller->category() == Decl::Class );
    
    if (caller != ObjectDecl) {
	
	vector< string > parameters;
	ClassDecl &super = static_cast< ClassDecl & >(*decl()->container());
	assert( super.category() == Decl::Class );
	
	parameters.push_back('(' + super.cType() + " *) " + MANGLE_THIS_VAR(+));
	args()->emitExpressionList( context, parameters );
	
	context << decl()->cMethodNameStatic() << '('
		<< delimit( parameters )
		<< ");" << endCline;
    }
    
    // The rules for chaining one constructor onto another
    // constructor for the same class ultimately mean that
    // only a subset of "terminal" constructors will call
    // the superclass constructor.  The same subset can also
    // bear responsibility for intializing instance fields
    // that were declared with initializer expressions.
    context << caller->cInitFieldsName() << "( " << MANGLE_THIS_VAR(<<) << " );" << endCline;
}

void StaticInitNode::codeGen( CodeContext &context )
{
  context.setDebugLine(this);

    // semantics??
    block()->codeGen( context );
}


void EmptyStmtNode::emitStatement( CodeContext & context)
{
}

void BlockNode::emitStatement( CodeContext &context )
{
  if (stmts()->arity() == 0)
    context << " ;" << endCline;
  else if (stmts()->arity() == 1)
    stmts()->codeGen( context );
  else {
    CodeContext subcontext( context );
    stmts()->codeGen( subcontext );
    subcontext.setDebugLine(this);
  } 
}

UniqueId gotoLabelGenerator("L");

static string *gotolabel(TreeNode *t) 
{
  static Map m;
  string *& result = m[t];
  if (result)
    return result;
  else
    return (result = new string(gotoLabelGenerator.next()));
}

void LabeledStmtNode::emitStatement( CodeContext &context )
{
  context << *gotolabel(this) << ": ;\n";
  stmt()->codeGen( context );
}

void GotoNode::emitStatement( CodeContext &context )
{
  context << "goto " << *gotolabel(destination()) << ";" << endCline;
}

void IfStmtNode::emitStatement( CodeContext &context )
{
    const string predicate = condition()->emitExpression( context );
    context << "if (" << predicate << ") ";
    {
	CodeContext thenContext( context );
	thenPart()->codeGen( thenContext );
    }
    if (!elsePart()->absent()) {
	context << "else ";
	CodeContext elseContext( context );
	elsePart()->codeGen( elseContext );
    }
}

void CodeLiteralNode::emitStatement ( CodeContext &context ) {
  context << codeString() << endCline;
}

void ThrowNode::emitStatement( CodeContext &context )
{
  string exception = expr()->emitExpression( context );  
  context << "THROW(" << exception << ");" << endCline;
}


void ParameterNode::codeGen( CodeContext & context)
{
  context.setDebugLine(this);

    // dont worry (handled by methodDeclNode)
}


void NameNode::codeGen( CodeContext &context )
{
  context.setDebugLine(this);

    if (!qualifier()->absent()) {
	qualifier()->codeGen( context );
	context << "_" << *ident();
    } else
	context << *ident();
}


void PartitionStmtNode::emitStatement( CodeContext & context)
{
}
