#include "AST.h"
#include "UniqueId.h"
#include "code.h"
#include "code-util.h"
#include "compiler.h"
#include "decls.h"
#include "delimit.h"
#include "optimize.h"
#include "lower.h"
#include "st-field.h"
#include "streq.h"

/*

The intermediate form (IF) we are constructing has a simplified form
compared to what the frontend creates and manipulates.

INPUT: a correct AST.  A correct AST has parent pointers set correctly
	and has types on all expressions, among other things.

OUTPUT: IF, a correct AST that conforms to the rules below.


All expressions must be "baby expressions" of the following forms:
(this should be expressed in terms of AST nodes rather than
pseudo code)

1.  var
2a. var.field
2b. var.i0.i1....ik.field	 (where i0, ..., ik are immutable)
3a. op var			 (UnaryArithNode, NotNode, ComplementNode)
3b. var op var			 (BinaryArithNode, ShiftNode,
				  EqualityNode, RelationNode,
				  BitwiseNode, but not LogCondNode)
3c. var.method(var, var, ...)	 (method call)
3d. foo.method(var, var, ...)	 (static method call (foo = var or ObjectNode))
3e. new type(var, var, ...)	 (allocation)
3f. new type(var, ...)[var]...   (array allocation)
4.  var[var]			 (array access)
5.  (type) var			 (CastNode)
6.  var instanceof type		 (InstanceOfNode)
7.  ibroadcast var from var	 (see note below)
8.  [var, ..., var]              (PointNode)
9.  [var : var, var : var, ...], and variants  (DomainNode)

In the above, var means a local variable, compiler generated
temporary, or constant (PrimitiveLitNode or CodeLiteralExprNode).  In
practice, it will often be a temporary variable that is written once
and read once, but not always.  In "baby expressions" that contain
multiple "vars," the "vars" may be evaluated in any order.  Rewriting
to enforce ordering constraints specified by the source language must
take place before/during the conversion to IF.

In Titanium, "broadcast foo from bar" has special evaluations rules.
In the IF, evaluation (on all procs) of the sending proc and
evaluation (on the sending proc) of the value to be sent must precede
the ibroadcast node.  We could have used the same AST node type
(broadcast), but by having two different node types there should be
less confusion.  The type of ibroadcast is (T, int) -> T, where T is
any type, so we can't rewrite it as a normal method call.

An assignment must be an AssignNode of the following form:

e1 = e2		(where e1 is assignable from e2 and both are baby expressions)

All control flow must be expressed using method calls or the following
node types:

IfStmtNode
ForEachStmtNode (with ForEachPairNode)
SwitchNode (with CaseNode and SwitchBranchNode)
LabeledStmtNode
GotoNode
ReturnNode
SynchronizedNode
TryStmtNode, TryNode, CatchNode, FinallyNode, ThrowNode
BlockNode

*/


// DEBUG_LOWERING is mainly used for things "of current interest"---not much.
#define DEBUG_LOWERING 0

#define DEBUG_ADD_STMTS 0
#define DEBUG_GOTO 0
#define DEBUG_LOWER_FOR_TO_WHILE 0
#define DEBUG_IFEXPRNODE 0
#define DEBUG_LOWER_STMT_TO_BLOCK 0
#define DEBUG_LOWERING_OPTS 0
#define DEBUG_OPASSIGN 0

#define SHOWTREE(s, t) \
    do { cout << s << '\n'; (t)->print(cout, 0); cout << '\n'; } while (0)

static void showlist(const char *label, const llist<TreeNode *> *l)
{
  int length = l->size();
  if (length == 0)
    cout << label << " (list is empty)\n\n";
  else {
    cout << label << " (" << length << "):\n\n";
    foreach_const (p, llist<TreeNode *>, *l) {
      (*p)->print(cout, 4);
      cout << "\n\n";
    }
  }
}

/////////////////////////////////////////////////////////////////////////////

#define GOTO(n) (new GotoNode(n))
#define LABELED(n, pos) (new LabeledStmtNode(TreeNode::omitted, (n), (pos)))
#define NAKEDLABEL() LABELED(TreeNode::omitted, NoSourcePosition)

/////////////////////////////////////////////////////////////////////////////

/* Return true if t is a local variable, parameter, constant, or a ThisNode. */
/* Print an error message if return value is false and p is true. */
bool verifyIsVar(const TreeNode *t, bool p)
{
  if (isLocalVar(t) || isFormalParam(t) || isConstant(t) || isThisNode(t) ||
      isCodeLiteralExprNode(t))
    return true;

  if (p) {
    t->error("Internal error: expecting local variable or method parameter");
    t->print(cout, 2);
    cout << endl;
  }
  return false;
}

/* Return true iff t is a local variable, parameter, constant, or a ThisNode. */
bool isVar(const TreeNode *t)
{
  return verifyIsVar(t, false);
}

// Walks over the tree to make sure there aren't any aliased nodes except
// for type nodes and TreeNode::omitted. The use/def edges are attached
// directly to ObjectNodes and thus aliasing is Uncool.
// Returns true if there is aliasing.
bool checkAliasing(TreeNode *n) {
  bool aliasFound = false;
  int childCount = n->arity();
  for (int sweep = 0; sweep < childCount; sweep++) {
    TreeNode *child = n->child(sweep);
    if (child != TreeNode::omitted && 
	!strstr(child->oper_name(), "Type")) {
      if (n->child(sweep)->parent() != n) {
	cout << "Aliasing error! Child is:\n";
	n->child(sweep)->print(cout,0);
	cout << "\nParent in downward direction is: " << endl;
	n->print(cout,0);
	cout << "\nParent in upward direction is:\n";
	n->child(sweep)->parent()->print(cout,0);
	cout << endl;
	aliasFound = true;
      }
      aliasFound |= checkAliasing(n->child(sweep));
    }
  }
  return aliasFound;
}


/////////////////////////////////////////////////////////////////////////////

static UniqueId tempGenerator( "x" );

void resetIFtemp()
{
  tempGenerator.reset();
}

static TreeNode *MakeTemporary(TreeNode *expr, TreeNode *&declStmt,
			       TreeNode *&assnStmt)
{
  SourcePosn p = expr->position();
  const string *str = intern(tempGenerator.next().c_str());
  TreeNode *name = new NameNode (TreeNode::omitted, str, (Decl *)NULL);
  TreeNode *varDecl =
    new VarDeclNode (false, expr->type()->addModifiers(Common::CompilerGenerated),
		     name, TreeNode::omitted, p);
  Decl *d = new LocalVarDecl(varDecl->simpName()->ident(),
			     varDecl->dtype(), varDecl, true);
  varDecl->simpName()->decl(d);
  declStmt = varDecl;
  
  TreeNode *varNode = new ObjectNode (new NameNode(TreeNode::omitted,
						   str, d, p), p);
  varNode->type(expr->type());
  assnStmt = new ExpressionStmtNode(new AssignNode(varNode, expr, p), p);
  
  return CloneTree(varNode);
}

BlockNode *AddBlockStmts(BlockNode *b, TreeNode *insertPt,
			 llist<TreeNode *> *newStmts)
{
  if (!newStmts)
    return b;

#if DEBUG_ADD_STMTS  
  cout << "Adding to block:\n";
  b->print(cout, 0);
  cout << "\ninsertPt:\n";
  if (insertPt == NULL) cout << "NULL";
  else insertPt->print(cout, 0);
  for (llist<TreeNode *> *s = newStmts; s; s = s->tail()) {
    cout << "\nTo be added:\n";
    s->front()->print(cout, 0);
  }
  cout << "\n";
#endif
    TreeListNode *s = b->stmts();
    int c = s->arity();
    llist<TreeNode *> *t = NULL;
    if (c != 0) {
	for (int i=c-1; i>=0; i--) {
	    t = cons(s->child(i), t);
	    if (s->child(i) == insertPt) {
		for ( ; newStmts; newStmts = newStmts->tail() ) {
		    t = cons(newStmts->front(), t);
		}
	    }
	}
    } else {
	for ( ; newStmts; newStmts = newStmts->tail() )
	    t = cons(newStmts->front(), t);
    }
    b->stmts(new TreeListNode(t));
#if DEBUG_ADD_STMTS  
  cout << "Result of adding to block:\n";
  b->print(cout, 0);
  cout << "\n";
#endif
    return b;
}

SwitchBranchNode *AddSwitchBranchStmts(SwitchBranchNode *b, TreeNode *insertPt,
				       llist<TreeNode *> *newStmts)
{
    TreeListNode *s = b->stmts();
    int c = s->arity();
    llist<TreeNode *> *t = NULL;
    if (c != 0) {
	for (int i=c-1; i>=0; i--) {
	    t = cons(s->child(i), t);
	    if (s->child(i) == insertPt) {
		for ( ; newStmts; newStmts = newStmts->tail() ) {
		    t = cons(newStmts->front(), t);
		}
	    }
	}
    } else {
	for ( ; newStmts; newStmts = newStmts->tail() )
	    t = cons(newStmts->front(), t);
    }
    b->stmts(new TreeListNode(t));
    return b;
}

void AddStmts(TreeNode *insertPt, llist<TreeNode *> *newStmts)
{
    TreeNode *container = insertPt->parent()->parent();
    if (newStmts)
      if (isBlockNode(container))
	AddBlockStmts((BlockNode *) container,
		      insertPt, newStmts);
      else if (isSwitchBranchNode(container))
	AddSwitchBranchStmts((SwitchBranchNode *) container,
			     insertPt, newStmts);
      else assert(0);
}

// newStmts is in reverse order...
BlockNode *AddBlockStmtsAtBegin(BlockNode *b, llist<TreeNode *> *newStmts)
{
#if DEBUG_ADD_STMTS  
  cout << "Adding to block with arity " << b->stmts()->arity();
  cout << "\n";
#endif
  TreeNode *insert = (b->stmts()->arity() ? b->stmts()->child(0) : (TreeNode *)NULL);
  return AddBlockStmts(b, insert, newStmts);
}

/* Return an expression for the default initializer for a given type. */
TreeNode *defaultInitializer(TypeNode *t,
			     const SourcePosn pos)
{
  TreeNode *result;
  if (t->isTitaniumArrayType() || t->isReference())
    result = new CastNode(new NullPntrNode(), t, pos);
  else if (t->isImmutable()) {
    result = new CodeLiteralExprNode(callNoArgConstructor(t), pos);
  } else if (t->isPrimitive())
    result = new CastNode(new PrimitiveLitNode((int32) 0), t, pos);
  else
    assert(0);
  result->type(t);
  return result;
}

/* Convert any stmt into a self-contained block that is lowered. */
/* Has no side effects except on t and its children. */
static BlockNode *stmt_to_lowered_block(TreeNode *t)
{
  t = new BlockNode(cons(t));

#if DEBUG_LOWER_STMT_TO_BLOCK
  static int count = 0;
  int n = count++;
  cout << "Lowering stmt " << n << " to block:\n";
  t->print(cout, 2);
  cout << "\n";
#endif

  /* At this point t is a block with one stmt in it. */

  llist<TreeNode *> *substmts = NULL;
  llist<TreeNode *> *subdecls = NULL;
  TreeNode *c = t->stmts()->child(0)->lower(subdecls, substmts);

#if DEBUG_LOWER_STMT_TO_BLOCK
  cout << "Lowered stmt " << n << " to block.  intermediate results:\n";
  cout << "c is:\n";
  c->print(cout, 2);
  cout << "\n";
  cout.flush();
#endif

  /* At this point t is a block with one or more stmts in it.
     Put the lowered stmt last in the block. */

  t->stmts()->child(t->stmts()->arity() - 1, c);

#if DEBUG_LOWER_STMT_TO_BLOCK
  cout << "Lowered stmt " << n << " to block.  intermediate results:\n";
  cout << "t is:\n";
  t->print(cout, 2);
  cout << "\n";
  cout.flush();
#endif

  t = AddBlockStmtsAtBegin((BlockNode *) t, substmts);
  t = AddBlockStmtsAtBegin((BlockNode *) t, subdecls);

#if DEBUG_LOWER_STMT_TO_BLOCK
  cout << "Lowered stmt " << n << " to block.  Result:\n";
  t->print(cout, 2);
  cout << "\n";
  cout.flush();
#endif

  SortVarDeclsInBlock((BlockNode *) t);
  return (BlockNode *) t;
}

// Pull out any cleanup operations associated with "oldJump", and
// arrange for them to be performed before "newJump".
//
// If there are indeed cleanup operations, then the result will be a
// BlockNode ending with "newJump".  If there are no cleanup
// operations, the result will just be "newJump" by itself.

static TreeNode *prependCleanups(TreeNode *newJump, TreeNode *oldJump)
{
  llist< TreeNode * > * const cleanups = oldJump->cleanups();
  if (cleanups) {
    TreeNode * const cleanupBlock = new BlockNode( cleanups, oldJump->position() );
    llist< TreeNode * > *sequence = 0;
    push( sequence, newJump );
    push( sequence, cleanupBlock );
    return new BlockNode( sequence, oldJump->position() );
  } else
    return newJump;
}

/////////////////////////////////////////////////////////////////////////////

static TreeNode *ProcessExpr(TreeNode *expr,
			     llist<TreeNode *> *&decls,
			     llist<TreeNode *> *&stmts)
{
    const int childCount = expr->arity();
    for (int sweep = 0; sweep < childCount; ++sweep)
	expr->child(sweep, expr->child(sweep)->lower(decls, stmts));

    TreeNode *d, *s, *newExpr;
    newExpr = MakeTemporary(expr, d, s);
    decls = cons(d, decls);
    stmts = cons(s, stmts);
    return newExpr;
}

#define ProcessAllSubExpr(e, decls, stmts) \
    ProcessSubExpr((e), 0, 0, (decls), (stmts))

static TreeNode *ProcessSubExpr(TreeNode *expr,
				int a,
				int b,
				llist<TreeNode *> *&decls,
				llist<TreeNode *> *&stmts)
{
  const int childCount = expr->arity();
  for (int sweep = 0; sweep < childCount; ++sweep) {
    expr->child(sweep, expr->child(sweep)->lower(decls, stmts));
    if ((a == b || a == sweep || b == sweep) && !isVar(expr->child(sweep))) {
      TreeNode *d, *s;
      expr->child(sweep, MakeTemporary(expr->child(sweep), d, s));
      decls = cons(d, decls);
      stmts = cons(s, stmts);
    }
  }

  return expr;
}


/////////////////////////////////////////////////////////////////////////////

static TreeNode *generic_lower(TreeNode *t,
			       llist<TreeNode *> *&decls,
			       llist<TreeNode *> *&stmts)
{
  const int a = t->arity();
  for (int i = 0; i < a; i++)
    if (!t->child(i)->absent() && !isVar(t->child(i)))
      t->child(i, t->child(i)->lower(decls, stmts));
  return t;
}

TreeNode *TreeNode::lower(llist<TreeNode *> *&decls,
			  llist<TreeNode *> *&stmts)
{
  return generic_lower(this, decls, stmts);
}

TreeNode *TemplateDeclNode::lower( llist< TreeNode * > *&, llist< TreeNode * > *& )
{
  return this;
}

TreeNode *ExpressionStmtNode::lower(llist<TreeNode *> *&decls,
						 llist<TreeNode *> *&stmts)
{
  assert(stmts == NULL);
  child(0, child(0)->lower(decls, stmts));
  if (stmts == NULL)
    return this;
  else {
    TreeNode *result = new BlockNode(extend(reverse(stmts),
					    cons((TreeNode *) this)));
    stmts = NULL;
    return result;
  }
}

TreeNode *ReturnNode::lower(llist<TreeNode *> *&decls,
					 llist<TreeNode *> *&stmts)
{
  assert(stmts == NULL);

  // completely evaluate return value
  if (!expr()->absent() && !isVar(expr())) {
    TreeNode *decl, *stmt;
    expr(MakeTemporary(expr()->lower(decls, stmts), decl, stmt));
    push(stmts, stmt);
    push(decls, decl);
  }

  // perform any last-minute cleanups
  if (cleanups()) {
    TreeNode *cleanBlock = new BlockNode(cleanups(), position());
    cleanBlock = cleanBlock->lower(decls, stmts);
    push(stmts, cleanBlock);
    cleanups(0);
  }
  
  AddStmts(this, stmts);  
  stmts = NULL;
  return this;
}

TreeNode *ThrowNode::lower(llist<TreeNode *> *&decls,
					llist<TreeNode *> *&stmts)
{
  assert(stmts == NULL);
  child(0, child(0)->lower(decls, stmts));
  AddStmts(this, stmts);
  stmts = NULL;
  return this;
}

TreeNode *MonitorFetchInstanceNode::lower(llist<TreeNode *> *&decls,
						       llist<TreeNode *> *&stmts)
{
  assert(stmts == NULL);
  expr(expr()->lower(decls, stmts));
  AddStmts(this, stmts);
  stmts = NULL;
  return this;
}

TreeNode *MethodDeclNode::lower(llist<TreeNode *> *&decls,
					     llist<TreeNode *> *&stmts)
{
  assert(decls == NULL);
  assert(stmts == NULL);
  if (!body()->absent()) {
    if (!isBlockNode(body()))
      body(new BlockNode(cons(body())));
    body( body()->lower(decls, stmts) );
    body( AddBlockStmtsAtBegin((BlockNode *)body(), stmts) );
    body( AddBlockStmtsAtBegin((BlockNode *)body(), decls) );
    decls = NULL; stmts = NULL;
  }    
  return this;
}

TreeNode *ConstructorDeclNode::lower(llist<TreeNode *> *&decls,
						  llist<TreeNode *> *&stmts)
{
#if 0
  cout << "Lowering constructor:\n";
  print(cout, 2);
  cout << "\n";
#endif
  assert(decls == NULL);
  assert(stmts == NULL);
  if (!body()->absent()) {
    if (!isBlockNode(body()))
      body(new BlockNode(cons(body())));
    constructorCall(stmt_to_lowered_block(constructorCall()));
    body( body()->lower(decls, stmts) );
    body( AddBlockStmtsAtBegin((BlockNode *)body(), stmts) );
    body( AddBlockStmtsAtBegin((BlockNode *)body(), decls) );
    decls = NULL; stmts = NULL;
  }    
#if 0
  cout << "Constructor lowered to:\n";
  print(cout, 2);
  cout << "\n";
#endif
  return this;
}

TreeNode *BlockNode::lower(llist<TreeNode *> *&,
			   llist<TreeNode *> *&)
{
  TreeListNode *s = stmts();
  int childCount = s->arity();
  static int k = 0;
  int n = k++;

  /* First, make sure all vardecls have initializers. */
  for (int i = childCount - 1; i >= 0; i--) {
    if (isVarDeclNode(s->child(i))) {
      // cout << "Child " << i << " is a var decl" << endl;
      if (s->child(i)->initExpr()->absent()) {
	s->child(i)->initExpr(defaultInitializer(s->child(i)->dtype(),
						 s->child(i)->position()));
	// cout << "Set initializer to ";
	// cout << PCODE(s->child(i)->initExpr());
	// cout << endl;
      }
    }
  }

  SortVarDeclsInBlock(this);
  s = stmts();
  childCount = s->arity();

  // cout << n << ": Doing a block with " << childCount << " children\n";
  for (int i = childCount - 1; i >= 0; i--) {
    if (isVarDeclNode(s->child(i))) {
      // cout << "Child " << i << " is a var decl" << endl;
    } else {
      // cout << "Lowering child " << i << " of a block\n";
      // cout << "BEFORE " << n << '.' << i << "\n\n";
      // s->child(i)->print(cout, 0);
      // cout << "\n\n";

      s->child(i, stmt_to_lowered_block(s->child(i)));

      // cout << "AFTER " << n << '.' << i << "\n\n";
      // s->child(i)->print(cout, 0);
      // cout << "\n\n";
      // assert(s->child(i)->checkIF());
    }
  }

  stmts(s);

  // cout << "Lowered block: " << endl;
  // cout << PCODE(this);
  // cout << endl;
  // assert(checkIF());
  // cout << n << ": Done\n";
  return this;
}

TreeNode *IfStmtNode::lower(llist<TreeNode *> *&decls,
					 llist<TreeNode *> *&stmts)
{
  if (!isBlockNode(thenPart()))
    thenPart(new BlockNode(cons(thenPart())));

  if (!isBlockNode(elsePart()))
    elsePart(new BlockNode(cons(elsePart())));

  assert(stmts == NULL);
  assert(decls == NULL);
  condition( condition()->lower(decls, stmts) );
  AddStmts(this, stmts);
  AddStmts(this, decls);
  stmts = NULL;
  decls = NULL;
  thenPart( thenPart()->lower(decls, stmts) );
  elsePart( elsePart()->lower(decls, stmts) );
  assert(stmts == NULL);
  assert(decls == NULL);

#if 0 /* DEBUG_LOWERING */
  cout << "Lowered if statement to:\n";
  print(cout, 0);
  cout << "\n";
#endif

  return this;
}

TreeNode *TryStmtNode::lower(llist<TreeNode *> *&decls,
					  llist<TreeNode *> *&stmts)
{
    const int childCount = arity();
    for (int sweep = 0; sweep < childCount; ++sweep)
	child( sweep, child( sweep )->lower(decls, stmts));

    return this;
}

/* 
Lowering of while, do, and for, break, continue: use GOTO.
*/

// In the subtree t, replace a break of statement l to a GOTO to breakTo.
// Also replace a continue of statement l to a GOTO to continueTo.
static TreeNode *replaceBreakContinue(TreeNode *t, TreeNode *l,
				      TreeNode *breakTo, TreeNode *continueTo,
				      int &count)
{
  if (isBreakNode(t))
    if (t->destination() == l) {
      ++count;
      return prependCleanups(GOTO(breakTo), t);
    } else
      return t;
  else if (continueTo != NULL && isContinueNode(t))
    if (t->destination() == l) {
      return prependCleanups(GOTO(continueTo), t);
    } else
      return t;
  else {
    const int childCount = t->arity();
    for (int sweep = 0; sweep < childCount; ++sweep)
      t->child(sweep, replaceBreakContinue(t->child(sweep), l,
					   breakTo, continueTo, count));
    return t;
  }
}
#define replaceBreak(t, l, b, k) replaceBreakContinue((t), (l), (b), NULL, (k))

TreeNode *ForEachPairNode::lower(llist<TreeNode *> *&decls,
				 llist<TreeNode *> *&stmts)
{
  return ProcessSubExpr(this, -1, 1, decls, stmts);
}

TreeNode *ForEachStmtNode::lower(llist<TreeNode *> *&decls,
				 llist<TreeNode *> *&stmts)
{
  assert(stmts == NULL);
  assert(decls == NULL);

  if (!isBlockNode(stmt()))
    stmt(new BlockNode(cons(stmt()), stmt()->position()));

  TreeNode *after = NAKEDLABEL();
  TreeNode *continueTo = LABELED(stmt(), stmt()->position());
  int count = 0;

  assert(arity() == 2);

  stmt(new BlockNode(cons(continueTo)));
  replaceBreakContinue(this, this, after, continueTo, count);
  vars((TreeListNode *)vars()->lower(decls, stmts));
  AddStmts(this, stmts);
  AddStmts(this, decls);
  stmts = NULL;
  decls = NULL;
  stmt(stmt_to_lowered_block(stmt()));
  return ((count == 0) ?
	  (TreeNode *)this :
	  new BlockNode(cons((TreeNode *) this, cons(after)), position()));
}

TreeNode *LabeledStmtNode::lower(llist<TreeNode *> *&decls,
					      llist<TreeNode *> *&stmts)
{
  TreeNode *after = NAKEDLABEL(), *s = stmt();
  int k = 0;
  stmt(replaceBreak(s, s, after, k)->lower(decls, stmts));
  return ((k == 0) ?
	  (TreeNode *)this :
	  new BlockNode(cons((TreeNode *) this, cons(after))));
}

TreeNode *WhileNode::lower(llist<TreeNode *> *&decls,
					llist<TreeNode *> *&stmts)
{
  // while (test) do S;   =>    { goto bot; top: S; bot: if (test) goto top; }

#if DEBUG_GOTO
  cout << "Lowering while loop:\n";
  print(cout, 0);
  cout << "\n";
#endif

  const SourcePosn pos = position();

  TreeNode *body = stmt_to_lowered_block(stmt());
  TreeNode *s = LABELED(body, pos);
  BlockNode *b =
    stmt_to_lowered_block(new IfStmtNode(test(), GOTO(s), TreeNode::omitted,
					 test()->position()));
#if DEBUG_GOTO
  cout << "Test is:\n";
  b->print(cout, 0);
  cout << "\n";
#endif
		   
#if DEBUG_GOTO
  cout << "Body is:\n";
  body->print(cout, 0);
  cout << "\n";
#endif
		   
  TreeNode *bot = LABELED(b, pos);
  TreeNode *after = NAKEDLABEL();
  int k = 0;
  replaceBreakContinue(body, this, after, bot, k);
  b = new BlockNode(k == 0 ? cons(bot) : cons(bot, cons(after)));

  b = AddBlockStmtsAtBegin(b, cons(s, cons((TreeNode *) GOTO(bot))));

#if DEBUG_GOTO
  cout << "\nwhile loop lowered to:\n";
  b->print(cout, 0);
  cout << "\n";
#endif

  return b;
}

TreeNode *DoNode::lower(llist<TreeNode *> *&decls,
				     llist<TreeNode *> *&stmts)
{
  // do S while (test)    =>    { label: S; if (test) goto label; }

#if DEBUG_GOTO
  cout << "Lowering do loop:\n";
  print(cout, 0);
  cout << "\n";
#endif

  const SourcePosn pos = position();

  TreeNode *body = stmt_to_lowered_block(stmt());
  TreeNode *s = LABELED(body, pos);
  BlockNode *b =
    stmt_to_lowered_block(new IfStmtNode(test(), GOTO(s), TreeNode::omitted,
					 test()->position()));
		   
  TreeNode *bot = LABELED(b, pos);
  TreeNode *after = NAKEDLABEL();
  int k = 0;
  replaceBreakContinue(body, this, after, bot, k);
  b = new BlockNode((k == 0) ? cons(bot) : cons(bot, cons(after)));

  b = AddBlockStmtsAtBegin(b, cons(s));

#if DEBUG_GOTO
  cout << "\ndo loop lowered to:\n";
  b-> print(cout, 0);
  cout << "\n";
#endif

  return b;
}

TreeNode *ForNode::lower(llist<TreeNode *> *&decls,
				      llist<TreeNode *> *&stmts)
{
  TreeNode *updateLabel = NAKEDLABEL(), *afterLabel = NAKEDLABEL();

#if DEBUG_LOWER_FOR_TO_WHILE
  cout << "Lowering for loop:\n";
  print(cout, 0);
  cout << "\n";
#endif

  // Create the loop body: the body of the "for" plus the update stmts.
  TreeListNode *updStmts = update();
  llist<TreeNode *> *updList = NULL;
  for (int c = updStmts->arity() - 1; c >= 0; c--)
    updList = cons(updStmts->child(c), updList);
  updList = cons(stmt(), cons(updateLabel, updList));
  TreeNode *loopBody = new BlockNode(updList);
  int k = 0;
  replaceBreakContinue(loopBody, this, afterLabel, updateLabel, k);

  // Construct a while loop with appropriate test, and body from above.
  TreeNode *whileLoop;
  if (test()->absent())
    whileLoop = new WhileNode(newBoolLit(true), loopBody);
  else
    whileLoop = new WhileNode(test(), loopBody);

  // Handle the init stmts.
  TreeListNode *initStmts = init();
  llist<TreeNode *> *t =
    (k == 0 ? cons(whileLoop) : cons(whileLoop, cons(afterLabel)));
  // Attach the initStmts to a new created enclosing block
  // with the new while loop as the last statement of the block.
  for (int c = initStmts->arity() - 1; c >= 0; c--)
    t = cons(initStmts->child(c), t);
    
  BlockNode *b = new BlockNode(t);
#if DEBUG_LOWER_FOR_TO_WHILE
  cout << "for loop lowered to while:\n";
  b-> print(cout, 0);
  cout << "\n";
#endif
  return b->lower(decls, stmts);
}


TreeNode *SwitchNode::lower(llist<TreeNode *> *&decls,
					 llist<TreeNode *> *&newStmts)
{
  assert(newStmts == NULL);
  expr( expr()->lower(decls, newStmts) );
  AddStmts(this, newStmts);
  newStmts = NULL;

  TreeListNode *swB = switchBlocks();
  const int childCount = swB->arity();
  for (int sweep = 0; sweep < childCount; ++sweep) {
    SwitchBranchNode *b = (SwitchBranchNode *)swB->child(sweep);

    SortVarDeclsInSwitch(b);
    TreeListNode *s = b->stmts();
    const int num = s->arity();

    for (int i = 0; i < num; i++)
      if (!isVarDeclNode(s->child(i)))
	s->child(i, stmt_to_lowered_block(s->child(i)));
  }

  int k = 0;
  TreeNode *after = NAKEDLABEL();
  replaceBreak(this, this, after, k);

  return
    ((k == 0) ? (TreeNode *)this : new BlockNode(cons((TreeNode *) this, cons(after))));
}

/////////////////////////////////////////////////////////////////////////////
// Basic unary exprs
/////////////////////////////////////////////////////////////////////////////

#define UNARY(this, decls, stmts) \
	ProcessSubExpr((this), 0, -1, (decls), (stmts))

TreeNode *UnaryArithNode::lower(llist<TreeNode *> *&decls,
				llist<TreeNode *> *&stmts)
{
  return UNARY(this, decls, stmts);
}

TreeNode *ComplementNode::lower(llist<TreeNode *> *&decls,
				llist<TreeNode *> *&stmts)
{
  return UNARY(this, decls, stmts);
}

TreeNode *NotNode::lower(llist<TreeNode *> *&decls,
			 llist<TreeNode *> *&stmts)
{
  return UNARY(this, decls, stmts);
}

TreeNode *InstanceOfNode::lower(llist<TreeNode *> *&decls,
				llist<TreeNode *> *&stmts)
{
  if (opnd0()->type()->typeIdent(dtype()))
    return newBoolLit(true, position());
  else
    return UNARY(this, decls, stmts);
}

TreeNode *CastNode::lower(llist<TreeNode *> *&decls,
			  llist<TreeNode *> *&stmts)
{
  if (opnd0()->type()->typeIdent(dtype()))
    return opnd0()->lower(decls, stmts);
  else
    return UNARY(this, decls, stmts);
}

TreeNode *ObjectFieldAccessNode::lower(llist<TreeNode *> *&decls,
				       llist<TreeNode *> *&stmts)
{
  if ((decl()->modifiers() & Static) &&
      (isThisNode(object()) || isObjectNode(object())))
    return this;
  else
    return UNARY(this, decls, stmts);
}

TreeNode *AllocateArrayDimensionNode::lower(llist<TreeNode *> *&decls,
					    llist<TreeNode *> *&stmts)
{
  return UNARY(this, decls, stmts);
}

extern bool isIFPointOrDomainConstructor(const TreeNode *t);

TreeNode *DomainNode::lower(llist<TreeNode *> *&decls,
			    llist<TreeNode *> *&stmts)
{
  if (isIFPointOrDomainConstructor(this))
    return this;
  // cout << PCODE(this); cout << " lowered to ";
  args()->child(0, ProcessAllSubExpr(args()->child(0), decls, stmts));
  // cout << PCODE(this); cout << endl;
  return this;
}

TreeNode *PointNode::lower(llist<TreeNode *> *&decls,
			   llist<TreeNode *> *&stmts)
{
  // cout << PCODE(this) << " lowered to ";
  if (!isIFPointOrDomainConstructor(this))
    args((TreeListNode *) ProcessAllSubExpr(args(), decls, stmts));
  // cout << PCODE(this) << endl;
  return this;
}

/////////////////////////////////////////////////////////////////////////////
// Basic binary exprs
/////////////////////////////////////////////////////////////////////////////

#define BINARY(this, decls, stmts) \
	ProcessSubExpr((this), 0, 1, (decls), (stmts))

TreeNode *BinaryArithNode::lower(llist<TreeNode *> *&decls,
				 llist<TreeNode *> *&stmts)
{
  return BINARY(this, decls, stmts);
}

TreeNode *ShiftNode::lower(llist<TreeNode *> *&decls,
			   llist<TreeNode *> *&stmts)
{
  return BINARY(this, decls, stmts);
}

TreeNode *EqualityNode::lower(llist<TreeNode *> *&decls,
			      llist<TreeNode *> *&stmts)
{
  return BINARY(this, decls, stmts);
}

TreeNode *RelationNode::lower(llist<TreeNode *> *&decls,
			      llist<TreeNode *> *&stmts)
{
  return BINARY(this, decls, stmts);
}

TreeNode *BitwiseNode::lower(llist<TreeNode *> *&decls,
			     llist<TreeNode *> *&stmts)
{
  return BINARY(this, decls, stmts);
}

TreeNode *CandNode::lower(llist<TreeNode *> *&decls,
			  llist<TreeNode *> *&stmts)
{
  TreeNode *n = new IfExprNode(opnd0(), opnd1(),
			       newBoolLit(false, position()), position());
  n->type(child(1)->type());
  return n->lower(decls, stmts);
}

TreeNode *CorNode::lower(llist<TreeNode *> *&decls,
			 llist<TreeNode *> *&stmts)
{
  TreeNode *n = new IfExprNode(opnd0(), newBoolLit(true, position()),
			       opnd1(), position());
  n->type(child(1)->type());
  return n->lower(decls, stmts);
}

TreeNode *ArrayAccessNode::lower(llist<TreeNode *> *&decls,
				 llist<TreeNode *> *&stmts)
{
  return BINARY(this, decls, stmts);
}

/////////////////////////////////////////////////////////////////////////////


TreeNode *AssignNode::lower(llist<TreeNode *> *&decls,
			    llist<TreeNode *> *&stmts)
{
  child(0, child(0)->lower(decls, stmts));
  child(1, child(1)->lower(decls, stmts));

  if (parent() != NULL && isExpressionStmtNode(parent()))
    return this;
  else {
    TreeNode *t = CloneTree(child(1));
    push( stmts, static_cast<TreeNode *>(new ExpressionStmtNode(this)) );
    return t;
  }
}

static TreeNode *opAssignNode_lower(TreeNode *x,
				    llist<TreeNode *> *&decls,
				    llist<TreeNode *> *&stmts)
{
  TreeNode *t, *d, *s;
  TreeNode *c1 = x->child(0)->lower(decls, stmts);
  TreeNode *cc = CloneTree(c1);
  TreeNode *c2 = x->child(1);
  const SourcePosn pos = x->position();

  if (isMultAssignNode(x)) {
    t = new MultNode(c1, c2, pos);
  } else if (isDivAssignNode(x)) {
    t = new DivNode(c1, c2, pos);
  } else if (isRemAssignNode(x)) {
    t = new RemNode(c1, c2, pos);
  } else if (isPlusAssignNode(x)) {
    t = new PlusNode(c1, c2, pos);
  } else if (isMinusAssignNode(x)) {
    t = new MinusNode(c1, c2, pos);	
  } else if (isLeftShiftLogAssignNode(x)) {
    t = new LeftShiftLogNode(c1, c2, pos);
  } else if (isRightShiftLogAssignNode(x)) {
    t = new RightShiftLogNode(c1, c2, pos);
  } else if (isRightShiftArithAssignNode(x)) {
    t = new RightShiftArithNode(c1, c2, pos);
  } else if (isBitAndAssignNode(x)) {
    t = new BitAndNode(c1, c2, pos);
  } else if (isBitXorAssignNode(x)) {
    t = new BitXorNode(c1, c2, pos);
  } else if (isBitOrAssignNode(x)) {
    t = new BitOrNode(c1, c2, pos);
  } else if (isMethodCallAssignNode(x)) {
    t = new MethodCallNode(c1, NULL, pos);
    t->child(1, c2);
    cc = cc->object();
  } else assert(0);

  // Is this correct?
  t->type(x->type());

  if (isExpressionStmtNode(x->parent())) {
    TreeNode *r = new AssignNode(cc, t, pos);
    r->type(cc->type());
    return r->lower(decls, stmts);
  } else {
    t = t->lower(decls, stmts);
    if (!isVar(t)) {
      t = MakeTemporary(t, d, s);
      stmts = cons(s, stmts);
      decls = cons(d, decls);
    }
    TreeNode *result = CloneTree(t);
    TreeNode *r = new AssignNode(cc, t, pos);
    r->type(cc->type());
    TreeNode *assign = r->lower(decls, stmts);
    stmts = cons(static_cast<TreeNode *>(new ExpressionStmtNode(assign, pos)),
		 stmts);
    return result;
  }
}

TreeNode *BinaryArithAssignNode::lower(llist<TreeNode *> *&decls,
				       llist<TreeNode *> *&stmts)
{
  return opAssignNode_lower(this, decls, stmts);
}

TreeNode *MethodCallAssignNode::lower(llist<TreeNode *> *&decls,
				      llist<TreeNode *> *&stmts)
{
  return opAssignNode_lower(this, decls, stmts);
}

TreeNode *ShiftAssignNode::lower(llist<TreeNode *> *&decls,
				 llist<TreeNode *> *&stmts)
{
  return opAssignNode_lower(this, decls, stmts);
}

TreeNode *BitwiseAssignNode::lower(llist<TreeNode *> *&decls,
				   llist<TreeNode *> *&stmts)
{
  return opAssignNode_lower(this, decls, stmts);
}

// See transform_init().
TreeNode *FieldDeclNode::lower(llist<TreeNode *> *&decls,
			       llist<TreeNode *> *&stmts)
{
  return this;
}

TreeNode *IfExprNode::lower(llist<TreeNode *> *&decls,
			    llist<TreeNode *> *&stmts)
{
  const SourcePosn pos = position();
  TreeNode *etrue = thenOpnd(), *efalse = elseOpnd();
  TypeNode *T = etrue->type();
  TreeNode *varNode, *a;

#if DEBUG_IFEXPRNODE
  cout << "Lowering a ? b : c\netrue = \n";
  etrue->print(cout, 0);
  cout << "\nefalse = \n";
  efalse->print(cout, 0);
  cout << "\n";
#endif

  // Create a variable that will contain the result;
  const string *str = intern(tempGenerator.next().c_str());
  TreeNode *name = new NameNode (TreeNode::omitted, str, (Decl *)NULL);
  TreeNode *varDecl =
    new VarDeclNode (false, T->addModifiers(Common::CompilerGenerated),
		     name, TreeNode::omitted);
  Decl *d = new LocalVarDecl(varDecl->simpName()->ident(), varDecl->dtype(),
			     varDecl, true);
  varDecl->simpName()->decl(d);

  TreeNode *atrue, *afalse; // Blocks that compute the result and store it

  varNode = new ObjectNode(new NameNode(TreeNode::omitted, str, d));
  varNode->type(T);
  a = new AssignNode(varNode, etrue, pos);
  a->type(T);
  atrue = stmt_to_lowered_block(new ExpressionStmtNode(a));

  varNode = new ObjectNode(new NameNode(TreeNode::omitted, str, d));
  varNode->type(T);
  a = new AssignNode(varNode, efalse, pos);
  a->type(T);
  afalse = stmt_to_lowered_block(new ExpressionStmtNode(a));
  
#if DEBUG_IFEXPRNODE
  cout << "\natrue = \n";
  atrue->print(cout, 0);
  cout << "\nafalse = \n";
  afalse->print(cout, 0);
  cout << "\n";
#endif

  TreeNode *cond = condition()->lower(decls, stmts);
  TreeNode *d1, *s1, *c = MakeTemporary(cond, d1, s1);
  decls = cons(d1, cons(varDecl, decls));
  TreeNode *i = new IfStmtNode(c, atrue, afalse, pos);
  stmts = cons(i, cons(s1, stmts));

#if DEBUG_IFEXPRNODE
  cout << "i = \n";
  i->print(cout, 0);
  cout << "\n";
#endif

  TreeNode *result = new ObjectNode (new NameNode(TreeNode::omitted, str, d));
  result->type(T);
  return result;
}


/* If x is of the form A.arity() for a grid A then rewrite it to the
   appropriate int literal (don't bother checking whether A is null).
   This is valid because x has already been lowered, so we need not
   worry about side effects of A.  If this function returns x (i.e.,
   does not rewrite) then the call is done and the result is put in a
   temporary.  (See MethodCallNode::lower(), below.).  If
   this function does not return x then the method has been rewritten
   and the returned value is assumed to be a lowered AST representing
   the appropriate result.  */
static TreeNode *method_rewrite(TreeNode *x)
{
  const TreeNode *m = x->method();
  Decl *d = m->simpName()->decl();
  if ((d->modifiers() & Common::Static) && isOFAN(m)) {
    const TypeNode *t = m->object()->type();
    if (t->isTitaniumArrayType() && isNameNode(m->simpName())) {
      MethodDecl *d = static_cast< MethodDecl * >( m->simpName()->decl() );
      if (streq(d->name()->c_str(), "arity")) {
#if DEBUG_LOWERING_OPTS
	cout << "\nRewriting call to grid method arity()\n\n";
	x->print(cout, 2);
	cout << "\n\n";
#endif
	return new PrimitiveLitNode(Literal((int32) t->tiArity()));
      }
    }
  }
  
  return x;
}


TreeNode *MethodCallNode::lower(llist<TreeNode *> *&decls,
				llist<TreeNode *> *&stmts)
{
  method(method()->lower(decls, stmts));
  args((TreeListNode *) ProcessAllSubExpr(args(), decls, stmts));

  Decl *dcl = method()->simpName()->decl();
  if (dcl->type()->returnType()->kind() != Common::VoidKind) {
    TreeNode *d, *s, *newExpr;
    if ((newExpr = method_rewrite(this)) != this)
      return newExpr;
    newExpr = MakeTemporary(this, d, s);
    decls = cons(d, decls);
    stmts = cons(s, stmts);
    return newExpr;
  } else
    return this;    
}

TreeNode *IncrDecrNode::lower(llist<TreeNode *> *&decls,
			      llist<TreeNode *> *&stmts)
{
    TreeNode *t = child(0)->lower(decls, stmts);
    TreeNode *d, *s, *old;

    old = MakeTemporary(t, d, s);
    decls = cons(d, decls);
    stmts = cons(s, stmts);

    TreeNode *oneNode = new PrimitiveLitNode ((int32) 1);
    TreeNode *assign;
    // Do the increment/decrement operation
    if (isPostIncrNode(this) || isPreIncrNode(this))
	assign = new AssignNode(CloneTree(t), new PlusNode(old, oneNode));
    else
	assign = new AssignNode(CloneTree(t), new MinusNode(old, oneNode));
    assign->type(t->type());

    // If the order does not matter...
    if (isExpressionStmtNode(parent()))
	return assign;

    // The order matters.  So store the assignment operation.
    push(stmts, static_cast<TreeNode *>(new ExpressionStmtNode(assign)));

    // Return the old or the new value
    if (isPostIncrNode(this) || isPostDecrNode(this))
	return CloneTree(old);
    else
	return CloneTree(t);
}

static TreeNode *nodeMYPROC()
{
  TreeNode *m = new CodeLiteralExprNode("MYPROC");
  m->type(theIntType);
  return m;
}

/* Return an ExprNode that compares p to MYPROC. */
static TreeNode *isMYPROC(TreeNode *p)
{
  TreeNode *r = new EQNode(p, nodeMYPROC());
  r->type(theBoolType);
  return r;
}

TreeNode *BroadcastNode::lower(llist<TreeNode *> *&decls,
			       llist<TreeNode *> *&stmts)
{
  const SourcePosn pos = position();
  TreeNode *d1, *s1, *d2, *s2;
  TreeNode *p = MakeTemporary(proc()->lower(decls, stmts),
			      d1, s1);
  llist<TreeNode *> *substmts = NULL;
  llist<TreeNode *> *subdecls = NULL;
  TreeNode *e =
    MakeTemporary(expr()->lower(subdecls, substmts),
		  d2, s2);

  s2 = new BlockNode(cons(s2), pos);
  s2 = AddBlockStmtsAtBegin((BlockNode *) s2, substmts);
  s2 = AddBlockStmtsAtBegin((BlockNode *) s2, subdecls);

  TreeNode *ifMyBcastComputeExpr =
    new IfStmtNode(isMYPROC(CloneTree(p)), s2, TreeNode::omitted, pos);

  decls = cons(d1, cons(d2, decls));
  stmts = cons(ifMyBcastComputeExpr, cons(s1, stmts));

  TreeNode *result = new IBroadcastNode(e, p, pos);
  result->type(type());
  return result;
}

TreeNode *AllocateNode::lower(llist<TreeNode *> *&decls,
					   llist<TreeNode *> *&stmts)
{
    return ProcessExpr(this, decls, stmts);
}

TreeNode *AllocateArrayNode::lower(llist<TreeNode *> *&decls,
						llist<TreeNode *> *&stmts)
{
  return ProcessExpr(this, decls, stmts);
}

/////////////////////////////////////////////////////////////////////////////

extern TreeNode *transform_init(TreeNode *t);

BlockNode *EncapsulateInits(vector<TreeNode *> inits)
{
    vector<TreeNode *>::iterator b = inits.begin();
    vector<TreeNode *>::iterator e = inits.end();
    llist<TreeNode *> *t = NULL;

    while (b < e) {
	t = cons(transform_init(*b++), t);
    }
    t = dreverse(t);
    BlockNode *block = new BlockNode(t);
    llist<TreeNode *> *d = NULL, *s = NULL;
    block->lower(d, s);
    block = AddBlockStmtsAtBegin(block, s);
    block = AddBlockStmtsAtBegin(block, d);
    return block;
}

TreeNode *transform_init(TreeNode *t)
{
    if (isStaticInitNode(t))
	return stmt_to_lowered_block(t);

    if (isFieldDeclNode(t)) {
        // if (t->initExpr()->absent())
        //   t->initExpr(defaultInitializer(t->dtype(), t->position()));
	if (t->initExpr()->absent())
	    return t;
	TypeNode *cl = t->decl()->container()->asType();
	TreeNode *sname = new NameNode(TreeNode::omitted,
				       t->decl()->name(),
				       t->decl());
	bool l = cl->isReference();
	TreeNode *fld = new ObjectFieldAccessNode(new ThisNode(cl, l), sname,
						  t->position());
	TreeNode *a = new AssignNode(fld, t->initExpr(),
				     t->position());
	a->type(t->initExpr()->type());
	return stmt_to_lowered_block(new ExpressionStmtNode(a));
				     
    }
    assert(0);
}


/////////////////////////////////////////////////////////////////////////////

TreeNode * TreeNode::collapseTrivialBlocks()
{
  const int childCount = arity();
  for (int sweep = 0; sweep < childCount; ++sweep)
    child(sweep, child( sweep )->collapseTrivialBlocks());
  
  return this;
}

TreeNode *TemplateDeclNode::collapseTrivialBlocks()
{
  return this;
}

TreeNode * MethodDeclNode::collapseTrivialBlocks()
{
  if (!body()->absent()) {
    /* Body should be a block, and keep it that way. */
    assert(isBlockNode(body()));
    body()->child(0, (body()->child(0)->collapseTrivialBlocks()));
  }
  return this;
}

TreeNode * ConstructorDeclNode::collapseTrivialBlocks()
{
  if (!body()->absent()) {
    /* Body should be a block, and keep it that way. */
    assert(isBlockNode(body()));
    body()->child(0, (body()->child(0)->collapseTrivialBlocks()));
  }
  return this;
}

TreeNode * StaticInitNode::collapseTrivialBlocks()
{
  if (!block()->absent()) {
    /* Body should be a block, and keep it that way. */
    assert(isBlockNode(block()));
    block()->child(0, (block()->child(0)->collapseTrivialBlocks()));
  }
  return this;
}

TreeNode * SynchronizedNode::collapseTrivialBlocks()
{
  if (!stmt()->absent()) {
    /* Body should be a block, and keep it that way. */
    assert(isBlockNode(stmt()));
    stmt()->child(0, (stmt()->child(0)->collapseTrivialBlocks()));
  }
  return this;
}

TreeNode * BlockNode::collapseTrivialBlocks()
{
  switch (child(0)->arity()) {
  case 0: return new EmptyStmtNode(position());
  case 1: return child(0)->child(0)->collapseTrivialBlocks();
  default: return TreeNode::collapseTrivialBlocks();
  }
}
