#include <split-c.h>
#include <am.h>

/**************************
 * INITIALIZATION/STATICS *
 *************************/

static handler_t packHandlerIdx, unpackAllHandlerIdx, unpackOnlyHandlerIdx;
static handler_t packReplyHandlerIdx, unpackReplyHandlerIdx;
static handler_t deleteHandlerIdx, nullReplyHandlerIdx, allocHandlerIdx,
                 allocReplyHandlerIdx;
static volatile int pack_spin, unpack_spin;

typedef void *(*pack_method_t)(void *, int *, void *);
typedef void *(*unpack_method_t)(void *, void *);

static void packHandler(void *token, void *copy_desc, int copy_desc_size, 
			int _packMethod, int remoteBuffer, int dummy2, 
			int dummy3);
static void packReplyHandler(void *token, void *array_data, 
			     int array_data_size, int dummy0, 
			     int dummy1, int dummy2, int dummy3);
static void unpackAllHandler(void *token, void *packedArrayData, int nBytes, 
			     int _unpackMethod, int copyDescSize, int dummy2, 
			     int dummy3);
static void unpackOnlyHandler(void *token, void *copyDesc, int copyDescSize,
			      int bufAddr, int _unpackMethod, int dummy2,
			      int dummy3);
static void unpackReplyHandler(void *token);
static void deleteHandler(void *token, int addr, int dummy1, int dummy2,
			  int dummy3);
static void allocHandler(void *token, int size, int dummy1, int dummy2,
			  int dummy3);
static void allocReplyHandler(void *token, int remAddr, int dummy1, 
			      int dummy2, int dummy3);
static void nullReplyHandler(void *token);

void bulkSetupHandlers() {
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &packHandlerIdx, packHandler));
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &unpackAllHandlerIdx, unpackAllHandler));
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &unpackOnlyHandlerIdx, unpackOnlyHandler));
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &packReplyHandlerIdx, packReplyHandler));
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &unpackReplyHandlerIdx, unpackReplyHandler));
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &deleteHandlerIdx, deleteHandler));
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &allocHandlerIdx, allocHandler));
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &allocReplyHandlerIdx, 
			   allocReplyHandler));
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &nullReplyHandlerIdx, nullReplyHandler));
}

/***************************
 * Miscellaneous handlers. *
 ***************************/

static void nullReplyHandler(void *token) {
  return;
}

static void deleteHandler(void *token, int addr, int dummy1, int dummy2,
			  int dummy3) {
  free((void *)addr);
  AM_Safe(AM_Reply0(token, nullReplyHandlerIdx));
}

/* Size is in bytes. */
static void allocHandler(void *token, int size, int dummy1, int dummy2,
			  int dummy3) {
  void *buf = (void *)malloc(size);
  AM_Safe(AM_Reply4(token, allocReplyHandlerIdx, (int)buf, 0, 0, 0));
}

static volatile void *remoteAllocBuf;

static void allocReplyHandler(void *token, int buf, int dummy1, int dummy2,
			      int dummy3) {
  remoteAllocBuf = (void *)buf;
}

/************************************
 * Get an array from a remote node. *
 ***********************************/

/* Call _packMethod to pack the array with copy_desc as its "copy descriptor",
   returning the packed data to remote address remoteBuffer. */
static void packHandler(void *token, void *copy_desc, int copy_desc_size,
			int _packMethod, int remoteBuffer, int dummy2, 
			int dummy3) {
  int array_data_len;
  void *array_data;
  double *copy_desc_copy;
  pack_method_t packMethod = (pack_method_t)_packMethod;
  /* A NULL buffer pointer tells the packer to allocate its own buffer. */
  copy_desc_copy = (double *)malloc(copy_desc_size);
  memcpy(copy_desc_copy, copy_desc, copy_desc_size); /* Align the data. */
  array_data = (*packMethod)(copy_desc_copy, &array_data_len, NULL); 
  free(copy_desc_copy);

  if (array_data_len <= AM_MaxMedium()) {
    /* Common, fast case. Send back the packed data with the reply. */
    AM_Safe(AM_ReplyI4(token, packReplyHandlerIdx, array_data, 
		       array_data_len, remoteBuffer, 0, 0, 0));
    free(array_data);
  }
  else {
    /* Slower case. Why doesn't the darned AM_ReplyXfer() work? */
    AM_Safe(AM_ReplyI4(token, packReplyHandlerIdx, array_data,
		       AM_MaxMedium(), remoteBuffer, 1, (int)array_data, 
		       array_data_len));
  }
}

/* The reply handler needs to pass this information back to the caller. */
static void *remoteBuffer;
static int dataSize;

/* Sets the spin flag to 1 if the array fit in the message reply and we're
   done; otherwise it sets the spinlock variable to be the remote address 
   of the packed data, which must be sent over with a libsplitc bulk 
   transfer. */
static void packReplyHandler(void *token, void *array_data, 
			     int array_data_size, int localBuffer,
			     int useSplitC, int remoteBuf, int dataSz) {
  memcpy((void *)localBuffer, array_data, array_data_size);
  if (!useSplitC) {
    pack_spin = 1;
  }
  else {
    /* Point to where we want to bulk-fetch the data from. Increment by
       AM_MaxMedium() because some of the data got sent already. */
    remoteBuffer = (void *)((int)remoteBuf + AM_MaxMedium());
    dataSize = dataSz - AM_MaxMedium();
    pack_spin = 2;
  }
}

void get_array(void *pack_method, void *copy_desc, int copy_desc_size, 
	       int tgt_node, void *buffer) {
  pack_spin = 0;
  /* A copy_desc is an object that contains the array descriptor. */
  AM_Safe(AM_RequestI4(__sc_endpoint, tgt_node, packHandlerIdx, 
		       copy_desc, copy_desc_size, (int)pack_method, 
		       (int)buffer, 0, 0));
  while (!pack_spin) { AM_Safe(AM_Poll(__sc_bundle)); }
  /* If the data was too big to be sent in the reply, use a Split-C call
     to do the job. */
  if (pack_spin != 1) {
    ___globPtrType remoteBufPtr = toglobal(tgt_node, remoteBuffer);
    bulk_read((void *)((int)buffer+AM_MaxMedium()), remoteBufPtr, dataSize);
    AM_Safe(AM_Request4(__sc_endpoint, tgt_node, deleteHandlerIdx, 
			(int)remoteBuffer-AM_MaxMedium(), 0, 0, 0));
  }
}

/************************************
 * Write an array to a remote node. *
 ***********************************/

/* The fastest way of doing things....take the data given and unpack it. */
static void unpackAllHandler(void *token, void *packedArrayData, int nBytes, 
		   int _unpackMethod, int copyDescSize, int dummy2, 
		   int dummy3) {
  unpack_method_t unpackMethod = (unpack_method_t)_unpackMethod;
  void *copyDesc = packedArrayData;
  void *arrayData = (void *)((int)packedArrayData + copyDescSize);
  double *arrayDataCopy = (double *)malloc(nBytes - copyDescSize);
  double *copyDescCopy = (double *)malloc(copyDescSize);
  memcpy(arrayDataCopy, arrayData, nBytes - copyDescSize); /* Align the data. */
  memcpy(copyDescCopy, copyDesc, copyDescSize); /* Align the data. */
  (*unpackMethod)(copyDesc, arrayDataCopy);
  free(arrayDataCopy);
  free(copyDescCopy);
  AM_Safe(AM_Reply0(token, unpackReplyHandlerIdx));
}

static void unpackReplyHandler(void *token) {
  unpack_spin = 1;
}

/* The data was provided by previous calls; just unpack it with the given
   descriptor. */
static void unpackOnlyHandler(void *token, void *copyDesc, int copyDescSize,
			      int bufAddr, int _unpackMethod, int dummy2,
			      int dummy3) {
  unpack_method_t unpackMethod = (unpack_method_t)_unpackMethod;
  (*unpackMethod)(copyDesc, (void *)bufAddr);
  free((void *)bufAddr);
  AM_Safe(AM_Reply0(token, unpackReplyHandlerIdx));
}
  

/* Send a contiguous array of data to a node, and have it get unpacked into
   a Titanium array. */
void put_array(void *unpack_method, void *copy_desc, int copy_desc_size,
	       void *array_data, int array_data_size, int tgt_node) {
  void *data;
  int data_size = copy_desc_size + array_data_size;
  unpack_spin = 0;

  /* Fast(er), hopefully common case. */
  if (data_size <= AM_MaxMedium()) {
    /* The remote end needs the descriptor and the array data. Since the
       data is small, copying the array data will be faster than
       sending two messages. */
    data = (void *) malloc(data_size);
    memcpy(data, copy_desc, copy_desc_size);
    /* All math is done in bytes. sizeof(void*) should be irrelevant. */
    memcpy((void *)((int)data + copy_desc_size), array_data, array_data_size);
    AM_Safe(AM_RequestI4(__sc_endpoint, tgt_node, unpackAllHandlerIdx, 
			 data, data_size, (int)unpack_method, 
			 copy_desc_size, 0, 0));
    while (!unpack_spin) { AM_Safe(AM_Poll(__sc_bundle)); }
    free(data);
  }
  else { /* Slow case. */
    ___globPtrType remotePtr;
    /* Allocate a buffer to hold the array data on the remote side. */
    remoteAllocBuf = NULL;
    AM_Safe(AM_Request4(__sc_endpoint, tgt_node, allocHandlerIdx, 
			array_data_size, 0, 0, 0));
    while (!remoteAllocBuf) { AM_Safe(AM_Poll(__sc_bundle)); }
    /* Transfer the data to the buffer with libsplitc. */
    remotePtr = toglobal(tgt_node, (void *)remoteAllocBuf);
    bulk_write(remotePtr, array_data, array_data_size);
    /* Tell the remote side to unpack the data. */
    AM_Safe(AM_RequestI4(__sc_endpoint, tgt_node, unpackOnlyHandlerIdx,
			 copy_desc, copy_desc_size, (int)remoteAllocBuf, 
			 (int)unpack_method, 0, 0));
    while (!unpack_spin) { AM_Safe(AM_Poll(__sc_bundle)); }
  }
}
