/* widen.cc: Definitions for member functions that add the conversions
   implied by type checking (mostly widening) */

#include "AST.h"
#include "compiler.h"
#include "decls.h"
#include "utils.h"
#include "optimize.h"

/* (Kludge:keep state in globals) */
/* current method */
static MethodDeclNode *currentMethod;

TreeNode *TreeNode::typeWiden(TreeNode *from, TypeNode *to) const
{
  if (from->type()->typeIdent(to)) return from;
  // cout << "typeWiden: from=" << PCODE(from) << " to=" << PCODE(to) << endl;
  ExprNode * const cast = new CastNode(from, to, position());
  cast->type(to);
  return cast;
}

void CompileUnitNode::widen()
{
  foriter (type, types()->allChildren(), ChildIter)
    (*type)->widen();
}

void TemplateDeclNode::widen()
{
}

void TreeNode::widen()
{
  foriter (p, allChildren(), ChildIter)
    (*p)->widen();
  _widen();
}

void MethodDeclNode::widen()
{
  currentMethod = this;
  TreeNode::widen();
  currentMethod = NULL;
}

void TreeNode::_widen()
{
}

void UnaryArithNode::_widen()
{
  opnd0(typeWiden(opnd0(), type()));
}

void BinaryArithNode::_widen()
{
  opnd0(typeWiden(opnd0(), type()));
  opnd1(typeWiden(opnd1(), type()));
}

void BinaryArithAssignNode::_widen()
{
  opnd1(typeWiden(opnd1(), arithPromoteType(opnd0()->type(), opnd1()->type())));
}

void ShiftNode::_widen()
{
  opnd0(typeWiden(opnd0(), type()));
  opnd1(typeWiden(opnd1(), type()));
}

void ShiftAssignNode::_widen()
{
  opnd1(typeWiden(opnd1(), arithPromoteType(opnd0()->type(), opnd1()->type())));
}

void RelationNode::_widen()
{
  TypeNode *rtype = arithPromoteType(opnd0()->type(), opnd1()->type());

  opnd0(typeWiden(opnd0(), rtype));
  opnd1(typeWiden(opnd1(), rtype));
}

void EqualityNode::_widen()
{
  TypeNode &type0 = *opnd0()->type();
  TypeNode &type1 = *opnd1()->type();
  TypeNode *rtype = 0;
  
  if (type0.isArithType())
    rtype = arithPromoteType(&type0, &type1);
  else if (type0.isReference()) {
    static TypeNode * const globalType = ObjectDecl->asType();
    static TypeNode * const localType = globalType->newModifiers(Local);
    const bool isLocal = type0.isLocal() && type1.isLocal();
    rtype = isLocal ? localType : globalType;
  }

  if (rtype) {
    opnd0(typeWiden(opnd0(), rtype));
    opnd1(typeWiden(opnd1(), rtype));
  }
}

void BitwiseNode::_widen()
{
  opnd0(typeWiden(opnd0(), type()));
  opnd1(typeWiden(opnd1(), type()));
}

void BitwiseAssignNode::_widen()
{
  opnd1(typeWiden(opnd1(), arithPromoteType(opnd0()->type(), opnd1()->type())));
}

void ArrayAccessNode::_widen()
{
  if (index()->type()->isPrimitive()) // only for java arrays
    index(typeWiden(index(), theIntType));
}

void ArrayInitNode::_widen()
{
  TypeNode * const elementType = type()->elementType();
  
  foriter (initializer, initializers()->allChildren(), ChildIter)
    *initializer = typeWiden(*initializer, elementType);
}

void TreeNode::methodWiden(TreeNode *args, Decl *decl)
{
  MethodTypeNode *method = (MethodTypeNode *) decl->type();

  foriter2 (argType, method->paramTypes()->allChildren(), 
	    arg, args->allChildren(),
	    ChildIter)
    *arg = typeWiden(*arg, (TypeNode *)*argType);
}

void TreeNode::methodWiden()
{
  TreeNode *mthod = method();
  Decl *mdecl = mthod->decl();
  Modifiers mflags = mdecl->modifiers();

  methodWiden(args(), mdecl);

  // Check dispatching object for a local->global conversion
  if (!(mflags & Static) && !(mflags & Local))
    {
      // must be an ObjectFieldAccessNode (after rewrite...)
      TypeNode *dispatchType = mthod->object()->type();
      Modifiers dflags = dispatchType->modifiers();

      if (dflags & Local)
	mthod->object
	  (typeWiden
	   (mthod->object(),
	    dispatchType->newModifiers((Modifiers)(dflags & ~Local))));
    }
}

void MethodCallNode::_widen()
{
  methodWiden();
}

void MethodCallAssignNode::_widen()
{
  methodWiden();
}

void AllocateNode::_widen()
{
  methodWiden(args(), decl());
}

void ThisConstructorCallNode::_widen()
{
  methodWiden(args(), decl());
}

void SuperConstructorCallNode::_widen()
{
  if (decl()) methodWiden(args(), decl());
}

void AllocateArrayNode::_widen()
{
  foriter (dimExpr, dimExprs()->allChildren(), TreeNode::ChildIter)
    if (!(*dimExpr)->expr()->type()->isRectDomainType())
      (*dimExpr)->expr(typeWiden((*dimExpr)->expr(), theIntType));
}

void ComplementNode::_widen()
{
  opnd0(typeWiden(opnd0(), type()));
}

void PlusNode::_widen()
{
  if (!type()->isStringType())
    {
      opnd0(typeWiden(opnd0(), type()));
      opnd1(typeWiden(opnd1(), type()));
    }
}

void IfExprNode::_widen()
{
  thenOpnd(typeWiden(thenOpnd(), type()));
  elseOpnd(typeWiden(elseOpnd(), type()));
}

void AssignNode::_widen()
{
  opnd1(typeWiden(opnd1(), type()));
}

void PlusAssignNode::_widen()
{
  if (!type()->isStringType())
    opnd1(typeWiden(opnd1(), arithPromoteType(opnd0()->type(), opnd1()->type())));
}

void PointNode::_widen()
{
  foriter (expr, args()->allChildren(), TreeNode::ChildIter)
    *expr = typeWiden(*expr, theIntType);
}

void DomainNode::_widen()
{
  // Handle both domain syntaxes
  // Could convert one into the other here, really.
  if (!(args()->arity() == 1 && args()->child(0)->child(0)->type()->isPointType()))
    foriter (range, args()->allChildren(), TreeNode::ChildIter)
    foriter (expr, (*range)->allChildren(), TreeNode::ChildIter)
    *expr = typeWiden(*expr, theIntType);
}

/* Statements */
void FieldDeclNode::_widen()
{
  if (!initExpr()->absent())
    initExpr(typeWiden(initExpr(), dtype()));
}

void VarDeclNode::_widen()
{
  if (!initExpr()->absent())
    initExpr(typeWiden(initExpr(), dtype()));
}

void SwitchNode::_widen()
{
  expr(typeWiden(expr(), theIntType));
}

void ReturnNode::_widen()
{
  if (currentMethod) {
    TypeNode *rtype = currentMethod->returnType();

    if (rtype->kind() != VoidKind)
      expr(typeWiden(expr(), rtype));
  }
}


void ThrowNode::_widen()
{
  // exceptions are always thrown as global references
  const TypeNode &type = *expr()->type();
  const Modifiers modifiers = (Modifiers) (type.modifiers() & ~Local);
  TypeNode * const globalType = type.newModifiers( modifiers );
  expr( typeWiden( expr(), globalType ) );
}


void BroadcastNode::_widen()
{
  // widen local references into global references
  expr( typeWiden( expr(), type() ) );
}


// The type conversions inherent in the xxxAssignNode nodes cannot be
// represented by CastNode's. Instead, given a op= b, two functions are provided:
//   - widenSource: the type to which a should be widened to perform 'a op b'
//   - castResult: the type to which 'a op b' should be cast before assignment to a
// In both cases, NULL is returned if no conversion is necessary.
// None of these casts require run-time checks.
// widenSource is also the type at which 'op' is computed

static TypeNode *typeWidenTo(TreeNode *from, TypeNode *to)
{
  if (from->type()->typeIdent(to)) return NULL;
  else return to;
}

TypeNode *TreeNode::_widenSource()
{
  return typeWidenTo(opnd0(), arithPromoteType(opnd0()->type(), opnd1()->type()));
}

TypeNode *TreeNode::_castResult()
{
  return typeWidenTo(arithPromoteType(opnd0()->type(), opnd1()->type()),
		     type());
}

TypeNode *TreeNode::widenSource()
{
  undefined("widenSource");
  return NULL;
}

TypeNode *TreeNode::castResult()
{
  undefined("castResult");
  return NULL;
}

TypeNode *MethodCallAssignNode::widenSource()
{
  return NULL;
}

TypeNode *MethodCallAssignNode::castResult()
{
  return typeWidenTo(method()->decl()->type()->returnType(),
		     method()->object()->type());
}

TypeNode *BinaryArithAssignNode::widenSource()
{
  return _widenSource();
}

TypeNode *BinaryArithAssignNode::castResult()
{
  return _castResult();
}

TypeNode *ShiftAssignNode::widenSource()
{
  return _widenSource();
}

TypeNode *ShiftAssignNode::castResult()
{
  return _castResult();
}

TypeNode *BitwiseAssignNode::widenSource()
{
  return _widenSource();
}

TypeNode *BitwiseAssignNode::castResult()
{
  return _castResult();
}

TypeNode *PlusAssignNode::widenSource()
{
  if (type()->isStringType())
    return NULL;
  else 
    return _widenSource();
}

TypeNode *PlusAssignNode::castResult()
{
  if (type()->isStringType())
    return NULL;
  else
    return _castResult();
}

